import asyncio
import os
import sys
import threading
import time
import traceback
from contextlib import asynccontextmanager
from datetime import datetime
from importlib import resources as importlib_resources
from typing import Any, Dict, List, Optional, Union

import pandas as pd
import psutil
from dateutil import parser
from fastapi import (
    Depends,
    FastAPI,
    Request,
    Response,
    WebSocket,
    WebSocketDisconnect,
    status,
)
from fastapi.concurrency import run_in_threadpool
from fastapi.logger import logger
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from google.protobuf.json_format import MessageToDict
from prometheus_client import Gauge, start_http_server
from pydantic import BaseModel

import feast
from feast import proto_json, utils
from feast.constants import DEFAULT_FEATURE_SERVER_REGISTRY_TTL
from feast.data_source import PushMode
from feast.errors import (
    FeastError,
    FeatureViewNotFoundException,
)
from feast.feast_object import FeastObject
from feast.permissions.action import WRITE, AuthzedAction
from feast.permissions.security_manager import assert_permissions
from feast.permissions.server.rest import inject_user_details
from feast.permissions.server.utils import (
    ServerType,
    init_auth_manager,
    init_security_manager,
    str_to_auth_manager_type,
)

# Define prometheus metrics
cpu_usage_gauge = Gauge(
    "feast_feature_server_cpu_usage", "CPU usage of the Feast feature server"
)
memory_usage_gauge = Gauge(
    "feast_feature_server_memory_usage", "Memory usage of the Feast feature server"
)


# TODO: deprecate this in favor of push features
class WriteToFeatureStoreRequest(BaseModel):
    feature_view_name: str
    df: dict
    allow_registry_cache: bool = True
    transform_on_write: bool = True


class PushFeaturesRequest(BaseModel):
    push_source_name: str
    df: dict
    allow_registry_cache: bool = True
    to: str = "online"
    transform_on_write: bool = True


class MaterializeRequest(BaseModel):
    start_ts: Optional[str] = None
    end_ts: Optional[str] = None
    feature_views: Optional[List[str]] = None
    disable_event_timestamp: bool = False


class MaterializeIncrementalRequest(BaseModel):
    end_ts: str
    feature_views: Optional[List[str]] = None


class GetOnlineFeaturesRequest(BaseModel):
    entities: Dict[str, List[Any]]
    feature_service: Optional[str] = None
    features: List[str] = []
    full_feature_names: bool = False


class GetOnlineDocumentsRequest(BaseModel):
    feature_service: Optional[str] = None
    features: List[str] = []
    full_feature_names: bool = False
    top_k: Optional[int] = None
    query: Optional[List[float]] = None
    query_string: Optional[str] = None
    api_version: Optional[int] = 1


class ChatMessage(BaseModel):
    role: str
    content: str


class ChatRequest(BaseModel):
    messages: List[ChatMessage]


class ReadDocumentRequest(BaseModel):
    file_path: str


class SaveDocumentRequest(BaseModel):
    file_path: str
    data: dict


def _get_features(
    request: Union[GetOnlineFeaturesRequest, GetOnlineDocumentsRequest],
    store: "feast.FeatureStore",
):
    if request.feature_service:
        feature_service = store.get_feature_service(
            request.feature_service, allow_cache=True
        )
        assert_permissions(
            resource=feature_service, actions=[AuthzedAction.READ_ONLINE]
        )
        features = feature_service  # type: ignore
    else:
        all_feature_views, all_on_demand_feature_views = (
            utils._get_feature_views_to_use(
                store.registry,
                store.project,
                request.features,
                allow_cache=True,
                hide_dummy_entity=False,
            )
        )
        for feature_view in all_feature_views:
            assert_permissions(
                resource=feature_view, actions=[AuthzedAction.READ_ONLINE]
            )
        for od_feature_view in all_on_demand_feature_views:
            assert_permissions(
                resource=od_feature_view, actions=[AuthzedAction.READ_ONLINE]
            )
        features = request.features  # type: ignore
    return features


def get_app(
    store: "feast.FeatureStore",
    registry_ttl_sec: int = DEFAULT_FEATURE_SERVER_REGISTRY_TTL,
):
    """
    Creates a FastAPI app that can be used to start a feature server.

    Args:
        store: The FeatureStore to use for serving features
        registry_ttl_sec: The TTL in seconds for the registry cache

    Returns:
        A FastAPI app

    Example:
        ```python
        from feast import FeatureStore

        store = FeatureStore(repo_path="feature_repo")
        app = get_app(store)
        ```

    The app provides the following endpoints:
    - `/get-online-features`: Get online features
    - `/retrieve-online-documents`: Retrieve online documents
    - `/push`: Push features to the feature store
    - `/write-to-online-store`: Write to the online store
    - `/health`: Health check
    - `/materialize`: Materialize features
    - `/materialize-incremental`: Materialize features incrementally
    - `/chat`: Chat UI
    - `/ws/chat`: WebSocket endpoint for chat
    MCP Support:
    - If MCP is enabled in feature server configuration, MCP endpoints will be added automatically
    """
    proto_json.patch()
    # Asynchronously refresh registry, notifying shutdown and canceling the active timer if the app is shutting down
    registry_proto = None
    shutting_down = False
    active_timer: Optional[threading.Timer] = None

    def stop_refresh():
        nonlocal shutting_down
        shutting_down = True
        if active_timer:
            active_timer.cancel()

    def async_refresh():
        if shutting_down:
            return

        store.refresh_registry()
        nonlocal registry_proto
        registry_proto = store.registry.proto()

        if registry_ttl_sec:
            nonlocal active_timer
            active_timer = threading.Timer(registry_ttl_sec, async_refresh)
            active_timer.start()

    @asynccontextmanager
    async def lifespan(app: FastAPI):
        await store.initialize()
        async_refresh()
        yield
        stop_refresh()
        await store.close()

    app = FastAPI(lifespan=lifespan)

    @app.post(
        "/get-online-features",
        dependencies=[Depends(inject_user_details)],
    )
    async def get_online_features(request: GetOnlineFeaturesRequest) -> Dict[str, Any]:
        # Initialize parameters for FeatureStore.get_online_features(...) call
        features = await run_in_threadpool(_get_features, request, store)

        read_params = dict(
            features=features,
            entity_rows=request.entities,
            full_feature_names=request.full_feature_names,
        )

        if store._get_provider().async_supported.online.read:
            response = await store.get_online_features_async(**read_params)  # type: ignore
        else:
            response = await run_in_threadpool(
                lambda: store.get_online_features(**read_params)  # type: ignore
            )

        # Convert the Protobuf object to JSON and return it
        response_dict = await run_in_threadpool(
            MessageToDict,
            response.proto,
            preserving_proto_field_name=True,
            float_precision=18,
        )
        return response_dict

    @app.post(
        "/retrieve-online-documents",
        dependencies=[Depends(inject_user_details)],
    )
    async def retrieve_online_documents(
        request: GetOnlineDocumentsRequest,
    ) -> Dict[str, Any]:
        logger.warning(
            "This endpoint is in alpha and will be moved to /get-online-features when stable."
        )
        # Initialize parameters for FeatureStore.retrieve_online_documents_v2(...) call
        features = await run_in_threadpool(_get_features, request, store)

        read_params = dict(features=features, query=request.query, top_k=request.top_k)
        if request.api_version == 2 and request.query_string is not None:
            read_params["query_string"] = request.query_string

        if request.api_version == 2:
            response = await run_in_threadpool(
                lambda: store.retrieve_online_documents_v2(**read_params)  # type: ignore
            )
        else:
            response = await run_in_threadpool(
                lambda: store.retrieve_online_documents(**read_params)  # type: ignore
            )

        # Convert the Protobuf object to JSON and return it
        response_dict = await run_in_threadpool(
            MessageToDict,
            response.proto,
            preserving_proto_field_name=True,
            float_precision=18,
        )
        return response_dict

    @app.post("/push", dependencies=[Depends(inject_user_details)])
    async def push(request: PushFeaturesRequest) -> None:
        df = pd.DataFrame(request.df)
        actions = []
        if request.to == "offline":
            to = PushMode.OFFLINE
            actions = [AuthzedAction.WRITE_OFFLINE]
        elif request.to == "online":
            to = PushMode.ONLINE
            actions = [AuthzedAction.WRITE_ONLINE]
        elif request.to == "online_and_offline":
            to = PushMode.ONLINE_AND_OFFLINE
            actions = WRITE
        else:
            raise ValueError(
                f"{request.to} is not a supported push format. Please specify one of these ['online', 'offline', 'online_and_offline']."
            )

        from feast.data_source import PushSource

        all_fvs = store.list_feature_views(
            allow_cache=request.allow_registry_cache
        ) + store.list_stream_feature_views(allow_cache=request.allow_registry_cache)
        fvs_with_push_sources = {
            fv
            for fv in all_fvs
            if (
                fv.stream_source is not None
                and isinstance(fv.stream_source, PushSource)
                and fv.stream_source.name == request.push_source_name
            )
        }

        for feature_view in fvs_with_push_sources:
            assert_permissions(resource=feature_view, actions=actions)

        push_params = dict(
            push_source_name=request.push_source_name,
            df=df,
            allow_registry_cache=request.allow_registry_cache,
            to=to,
            transform_on_write=request.transform_on_write,
        )

        should_push_async = (
            store._get_provider().async_supported.online.write
            and to in [PushMode.ONLINE, PushMode.ONLINE_AND_OFFLINE]
        )
        if should_push_async:
            await store.push_async(**push_params)
        else:
            store.push(**push_params)

    def _get_feast_object(
        feature_view_name: str, allow_registry_cache: bool
    ) -> FeastObject:
        try:
            return store.get_stream_feature_view(  # type: ignore
                feature_view_name, allow_registry_cache=allow_registry_cache
            )
        except FeatureViewNotFoundException:
            return store.get_feature_view(  # type: ignore
                feature_view_name, allow_registry_cache=allow_registry_cache
            )

    @app.post("/write-to-online-store", dependencies=[Depends(inject_user_details)])
    def write_to_online_store(request: WriteToFeatureStoreRequest) -> None:
        df = pd.DataFrame(request.df)
        feature_view_name = request.feature_view_name
        allow_registry_cache = request.allow_registry_cache
        resource = _get_feast_object(feature_view_name, allow_registry_cache)
        assert_permissions(resource=resource, actions=[AuthzedAction.WRITE_ONLINE])
        store.write_to_online_store(
            feature_view_name=feature_view_name,
            df=df,
            allow_registry_cache=allow_registry_cache,
            transform_on_write=request.transform_on_write,
        )

    @app.get("/health")
    async def health():
        return (
            Response(status_code=status.HTTP_200_OK)
            if registry_proto
            else Response(status_code=status.HTTP_503_SERVICE_UNAVAILABLE)
        )

    @app.post("/chat")
    async def chat(request: ChatRequest):
        # Process the chat request
        # For now, just return dummy text
        return {"response": "This is a dummy response from the Feast feature server."}

    @app.post("/read-document")
    async def read_document_endpoint(request: ReadDocumentRequest):
        try:
            import os

            if not os.path.exists(request.file_path):
                return {"error": f"File not found: {request.file_path}"}

            with open(request.file_path, "r", encoding="utf-8") as file:
                content = file.read()

            return {"content": content, "file_path": request.file_path}
        except Exception as e:
            return {"error": str(e)}

    @app.post("/save-document")
    async def save_document_endpoint(request: SaveDocumentRequest):
        try:
            import json
            import os
            from pathlib import Path

            file_path = Path(request.file_path).resolve()
            if not str(file_path).startswith(os.getcwd()):
                return {"error": "Invalid file path"}

            base_name = file_path.stem
            labels_file = file_path.parent / f"{base_name}-labels.json"

            with open(labels_file, "w", encoding="utf-8") as file:
                json.dump(request.data, file, indent=2, ensure_ascii=False)

            return {"success": True, "saved_to": str(labels_file)}
        except Exception as e:
            return {"error": str(e)}

    @app.get("/chat")
    async def chat_ui():
        # Serve the chat UI
        static_dir_ref = importlib_resources.files(__spec__.parent) / "static/chat"  # type: ignore[name-defined, arg-type]
        with importlib_resources.as_file(static_dir_ref) as static_dir:
            with open(os.path.join(static_dir, "index.html")) as f:
                content = f.read()
        return Response(content=content, media_type="text/html")

    @app.post("/materialize", dependencies=[Depends(inject_user_details)])
    def materialize(request: MaterializeRequest) -> None:
        for feature_view in request.feature_views or []:
            assert_permissions(
                resource=_get_feast_object(feature_view, True),
                actions=[AuthzedAction.WRITE_ONLINE],
            )

        if request.disable_event_timestamp:
            # Query all available data and use current datetime as event timestamp
            now = datetime.now()
            start_date = datetime(
                1970, 1, 1
            )  # Beginning of time to capture all historical data
            end_date = now
        else:
            if not request.start_ts or not request.end_ts:
                raise ValueError(
                    "start_ts and end_ts are required when disable_event_timestamp is False"
                )
            start_date = utils.make_tzaware(parser.parse(request.start_ts))
            end_date = utils.make_tzaware(parser.parse(request.end_ts))

        store.materialize(
            start_date,
            end_date,
            request.feature_views,
            disable_event_timestamp=request.disable_event_timestamp,
        )

    @app.post("/materialize-incremental", dependencies=[Depends(inject_user_details)])
    def materialize_incremental(request: MaterializeIncrementalRequest) -> None:
        for feature_view in request.feature_views or []:
            assert_permissions(
                resource=_get_feast_object(feature_view, True),
                actions=[AuthzedAction.WRITE_ONLINE],
            )
        store.materialize_incremental(
            utils.make_tzaware(parser.parse(request.end_ts)), request.feature_views
        )

    @app.exception_handler(Exception)
    async def rest_exception_handler(request: Request, exc: Exception):
        # Print the original exception on the server side
        logger.exception(traceback.format_exc())

        if isinstance(exc, FeastError):
            return JSONResponse(
                status_code=exc.http_status_code(),
                content=exc.to_error_detail(),
            )
        else:
            return JSONResponse(
                status_code=500,
                content=str(exc),
            )

    # Chat WebSocket connection manager
    class ConnectionManager:
        def __init__(self):
            self.active_connections: List[WebSocket] = []

        async def connect(self, websocket: WebSocket):
            await websocket.accept()
            self.active_connections.append(websocket)

        def disconnect(self, websocket: WebSocket):
            self.active_connections.remove(websocket)

        async def send_message(self, message: str, websocket: WebSocket):
            await websocket.send_text(message)

    manager = ConnectionManager()

    @app.websocket("/ws/chat")
    async def websocket_endpoint(websocket: WebSocket):
        await manager.connect(websocket)
        try:
            while True:
                message = await websocket.receive_text()
                # Process the received message (currently unused but kept for future implementation)
                # For now, just return dummy text
                response = f"You sent: '{message}'. This is a dummy response from the Feast feature server."

                # Stream the response word by word
                words = response.split()
                for word in words:
                    await manager.send_message(word + " ", websocket)
                    await asyncio.sleep(0.1)  # Add a small delay between words
        except WebSocketDisconnect:
            manager.disconnect(websocket)

    # Mount static files
    static_dir_ref = importlib_resources.files(__spec__.parent) / "static"  # type: ignore[name-defined, arg-type]
    with importlib_resources.as_file(static_dir_ref) as static_dir:
        app.mount("/static", StaticFiles(directory=static_dir), name="static")

    # Add MCP support if enabled in feature server configuration
    _add_mcp_support_if_enabled(app, store)

    return app


def _add_mcp_support_if_enabled(app, store: "feast.FeatureStore"):
    """Add MCP support to the FastAPI app if enabled in configuration."""
    try:
        # Check if MCP is enabled in feature server config
        if (
            store.config.feature_server
            and hasattr(store.config.feature_server, "type")
            and store.config.feature_server.type == "mcp"
            and getattr(store.config.feature_server, "mcp_enabled", False)
        ):
            from feast.infra.mcp_servers.mcp_server import add_mcp_support_to_app

            mcp_server = add_mcp_support_to_app(app, store, store.config.feature_server)

            if mcp_server:
                logger.info("MCP support has been enabled for the Feast feature server")
            else:
                logger.warning("MCP support was requested but could not be enabled")
        else:
            logger.debug("MCP support is not enabled in feature server configuration")
    except Exception as e:
        logger.error(f"Error checking/adding MCP support: {e}")
        # Don't fail the entire server if MCP fails to initialize


if sys.platform != "win32":
    import gunicorn.app.base

    class FeastServeApplication(gunicorn.app.base.BaseApplication):
        def __init__(self, store: "feast.FeatureStore", **options):
            self._app = get_app(
                store=store,
                registry_ttl_sec=options["registry_ttl_sec"],
            )
            self._options = options
            super().__init__()

        def load_config(self):
            for key, value in self._options.items():
                if key.lower() in self.cfg.settings and value is not None:
                    self.cfg.set(key.lower(), value)

            self.cfg.set("worker_class", "uvicorn_worker.UvicornWorker")

        def load(self):
            return self._app


def monitor_resources(self, interval: int = 5):
    """Function to monitor and update CPU and memory usage metrics."""
    logger.debug(f"Starting resource monitoring with interval {interval} seconds")
    p = psutil.Process()
    logger.debug(f"PID is {p.pid}")
    while True:
        with p.oneshot():
            cpu_usage = p.cpu_percent()
            memory_usage = p.memory_percent()
            logger.debug(f"CPU usage: {cpu_usage}%, Memory usage: {memory_usage}%")
            logger.debug(f"CPU usage: {cpu_usage}%, Memory usage: {memory_usage}%")
            cpu_usage_gauge.set(cpu_usage)
            memory_usage_gauge.set(memory_usage)
        time.sleep(interval)


def start_server(
    store: "feast.FeatureStore",
    host: str,
    port: int,
    no_access_log: bool,
    workers: int,
    keep_alive_timeout: int,
    registry_ttl_sec: int,
    tls_key_path: str,
    tls_cert_path: str,
    metrics: bool,
):
    if (tls_key_path and not tls_cert_path) or (not tls_key_path and tls_cert_path):
        raise ValueError(
            "Both key and cert file paths are required to start server in TLS mode."
        )
    if metrics:
        logger.info("Starting Prometheus Server")
        start_http_server(8000)

        logger.debug("Starting background thread to monitor CPU and memory usage")
        monitoring_thread = threading.Thread(
            target=monitor_resources, args=(5,), daemon=True
        )
        monitoring_thread.start()

    logger.debug("start_server called")
    auth_type = str_to_auth_manager_type(store.config.auth_config.type)
    logger.info(f"Auth type: {auth_type}")
    init_security_manager(auth_type=auth_type, fs=store)
    logger.debug("Security manager initialized successfully")
    init_auth_manager(
        auth_type=auth_type,
        server_type=ServerType.REST,
        auth_config=store.config.auth_config,
    )
    logger.debug("Auth manager initialized successfully")

    if sys.platform != "win32":
        options = {
            "bind": f"{host}:{port}",
            "accesslog": None if no_access_log else "-",
            "workers": workers,
            "keepalive": keep_alive_timeout,
            "registry_ttl_sec": registry_ttl_sec,
        }

        # Add SSL options if the paths exist
        if tls_key_path and tls_cert_path:
            options["keyfile"] = tls_key_path
            options["certfile"] = tls_cert_path
        FeastServeApplication(store=store, **options).run()
    else:
        import uvicorn

        app = get_app(store, registry_ttl_sec)
        if tls_key_path and tls_cert_path:
            uvicorn.run(
                app,
                host=host,
                port=port,
                access_log=(not no_access_log),
                ssl_keyfile=tls_key_path,
                ssl_certfile=tls_cert_path,
            )
        else:
            uvicorn.run(app, host=host, port=port, access_log=(not no_access_log))



from typing import Literal

from feast.infra.feature_servers.base_config import BaseFeatureServerConfig


class LocalFeatureServerConfig(BaseFeatureServerConfig):
    # Feature server type selector.
    type: Literal["local"] = "local"

    # The endpoint definition for transformation_service
    transformation_service_endpoint: str = "localhost:6569"



from datetime import datetime
from typing import Optional

from feast.data_source import DataSource
from feast.infra.compute_engines.dag.context import ColumnInfo, ExecutionContext
from feast.infra.offline_stores.offline_store import RetrievalJob


def create_offline_store_retrieval_job(
    data_source: DataSource,
    column_info: ColumnInfo,
    context: ExecutionContext,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> RetrievalJob:
    """
    Create a retrieval job for the offline store.
    Args:
        data_source: The data source to pull from.
        column_info: Column information containing join keys, feature columns, and timestamps.
        context:
        start_time:
        end_time:

    Returns:

    """
    offline_store = context.offline_store
    # 📥 Reuse Feast's robust query resolver
    retrieval_job = offline_store.pull_all_from_table_or_query(
        config=context.repo_config,
        data_source=data_source,
        join_key_columns=column_info.join_keys,
        feature_name_columns=column_info.feature_cols,
        timestamp_field=column_info.ts_col,
        created_timestamp_column=column_info.created_ts_col,
        start_date=start_time,
        end_date=end_time,
    )
    return retrieval_job



import logging
import os

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def configure_ca_trust_store_env_variables():
    """
    configures the environment variable so that other libraries or servers refer to the TLS ca file path.
    :param ca_file_path:
    :return:
    """
    if (
        "FEAST_CA_CERT_FILE_PATH" in os.environ
        and os.environ["FEAST_CA_CERT_FILE_PATH"]
    ):
        logger.info(
            f"Feast CA Cert file path found in environment variable FEAST_CA_CERT_FILE_PATH={os.environ['FEAST_CA_CERT_FILE_PATH']}. Going to refer this path."
        )
        os.environ["SSL_CERT_FILE"] = os.environ["FEAST_CA_CERT_FILE_PATH"]
        os.environ["REQUESTS_CA_BUNDLE"] = os.environ["FEAST_CA_CERT_FILE_PATH"]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/Project.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x66\x65\x61st/core/Project.proto\x12\nfeast.core\x1a\x1fgoogle/protobuf/timestamp.proto\"W\n\x07Project\x12%\n\x04spec\x18\x01 \x01(\x0b\x32\x17.feast.core.ProjectSpec\x12%\n\x04meta\x18\x02 \x01(\x0b\x32\x17.feast.core.ProjectMeta\"\x9d\x01\n\x0bProjectSpec\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12/\n\x04tags\x18\x03 \x03(\x0b\x32!.feast.core.ProjectSpec.TagsEntry\x12\r\n\x05owner\x18\x04 \x01(\t\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x80\x01\n\x0bProjectMeta\x12\x35\n\x11\x63reated_timestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12:\n\x16last_updated_timestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampBQ\n\x10\x66\x65\x61st.proto.coreB\x0cProjectProtoZ/github.com/feast-dev/feast/go/protos/feast/coreb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.core.Project_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020feast.proto.coreB\014ProjectProtoZ/github.com/feast-dev/feast/go/protos/feast/core'
  _globals['_PROJECTSPEC_TAGSENTRY']._options = None
  _globals['_PROJECTSPEC_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_PROJECT']._serialized_start=73
  _globals['_PROJECT']._serialized_end=160
  _globals['_PROJECTSPEC']._serialized_start=163
  _globals['_PROJECTSPEC']._serialized_end=320
  _globals['_PROJECTSPEC_TAGSENTRY']._serialized_start=277
  _globals['_PROJECTSPEC_TAGSENTRY']._serialized_end=320
  _globals['_PROJECTMETA']._serialized_start=323
  _globals['_PROJECTMETA']._serialized_end=451
# @@protoc_insertion_point(module_scope)



from typing import List, Optional, Set

from feast.feature_view import FeatureView
from feast.infra.compute_engines.algorithms.topo import topological_sort
from feast.infra.compute_engines.dag.context import ExecutionContext
from feast.infra.compute_engines.dag.node import DAGNode
from feast.infra.compute_engines.dag.value import DAGValue


class FeatureViewNode(DAGNode):
    """
    Logical representation of a node in the FeatureView dependency DAG.
    """

    def __init__(
        self, view: FeatureView, inputs: Optional[List["FeatureViewNode"]] = None
    ):
        super().__init__(name=view.name)
        self.view: FeatureView = view
        self.inputs: List["FeatureViewNode"] = inputs or []  # type: ignore

    def execute(self, context: ExecutionContext) -> DAGValue:
        raise NotImplementedError(
            f"FeatureViewNode '{self.name}' does not implement execute method."
        )


class FeatureResolver:
    """
    Resolves FeatureViews into a dependency graph (DAG) of FeatureViewNode objects.
    This graph represents the logical dependencies between FeatureViews, allowing
    for ordered execution and cycle detection.
    """

    def __init__(self):
        self._visited: Set[str] = set()
        self._resolution_path: List[str] = []
        self._node_cache: dict[str, FeatureViewNode] = {}

    def resolve(self, feature_view: FeatureView) -> FeatureViewNode:
        """
        Entry point for resolving a FeatureView into a DAG node.

        Args:
            feature_view: The root FeatureView to build the dependency graph from.

        Returns:
            A FeatureViewNode representing the root of the logical dependency DAG.
        """
        return self._walk(feature_view)

    def _walk(self, view: FeatureView):
        """
        Recursive traversal of the FeatureView graph.

        If `source_view` is set on the FeatureView, a parent node is created and added.
        Cycles are detected using the visited set.

        Args:
            view: The FeatureView to process.
        """
        if view.name in self._resolution_path:
            cycle = " → ".join(self._resolution_path + [view.name])
            raise ValueError(f"Cycle detected in FeatureView DAG: {cycle}")

        if view.name in self._node_cache:
            return self._node_cache[view.name]

        node = FeatureViewNode(view)
        self._node_cache[view.name] = node

        self._resolution_path.append(view.name)
        if view.source_views:
            for upstream_view in view.source_views:
                input_node = self._walk(upstream_view)
                node.inputs.append(input_node)
        self._resolution_path.pop()

        return node

    def topological_sort(self, root: FeatureViewNode) -> List[FeatureViewNode]:
        return topological_sort(root)  # type: ignore

    def debug_dag(self, node: FeatureViewNode, depth=0):
        """
        Prints the FeatureView dependency DAG for debugging.

        Args:
            node: The root node to print from.
            depth: Internal argument used for recursive indentation.
        """
        indent = "  " * depth
        print(f"{indent}- {node.view.name}")
        for input_node in node.inputs:
            self.debug_dag(input_node, depth + 1)  # type: ignore



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import json
import os
import uuid
from pathlib import Path
from tempfile import TemporaryFile
from typing import Optional
from urllib.parse import urlparse

from feast.infra.registry.registry import RegistryConfig
from feast.infra.registry.registry_store import RegistryStore
from feast.protos.feast.core.Registry_pb2 import Registry as RegistryProto
from feast.utils import _utc_now

REGISTRY_SCHEMA_VERSION = "1"


class AzBlobRegistryStore(RegistryStore):
    def __init__(self, registry_config: RegistryConfig, repo_path: Path):
        try:
            import logging

            from azure.identity import DefaultAzureCredential
            from azure.storage.blob import BlobServiceClient
        except ImportError as e:
            from feast.errors import FeastExtrasDependencyImportError

            raise FeastExtrasDependencyImportError("az", str(e))

        self._uri = urlparse(registry_config.path)
        self._account_url = self._uri.scheme + "://" + self._uri.netloc
        container_path = self._uri.path.lstrip("/").split("/")
        self._container = container_path.pop(0)
        self._path = "/".join(container_path)

        try:
            # turn the verbosity of the blob client to warning and above (this reduces verbosity)
            logger = logging.getLogger("azure")
            logger.setLevel(logging.ERROR)

            # Attempt to use shared account key to login first
            if "REGISTRY_BLOB_KEY" in os.environ:
                client = BlobServiceClient(
                    account_url=self._account_url,
                    credential=os.environ["REGISTRY_BLOB_KEY"],
                )
                self.blob = client.get_blob_client(
                    container=self._container, blob=self._path
                )
                return

            default_credential = DefaultAzureCredential(
                exclude_shared_token_cache_credential=True
            )

            client = BlobServiceClient(
                account_url=self._account_url, credential=default_credential
            )
            self.blob = client.get_blob_client(
                container=self._container, blob=self._path
            )
        except Exception as e:
            print(
                f"Could not connect to blob. Check the following\nIs the URL specified correctly?\nIs you IAM role set to Storage Blob Data Contributor? \n Errored out with exception {e}"
            )

        return

    def get_registry_proto(self):
        file_obj = TemporaryFile()
        registry_proto = RegistryProto()

        if self.blob.exists():
            download_stream = self.blob.download_blob()
            file_obj.write(download_stream.readall())

            file_obj.seek(0)
            registry_proto.ParseFromString(file_obj.read())
            return registry_proto
        raise FileNotFoundError(
            f'Registry not found at path "{self._uri.geturl()}". Have you run "feast apply"?'
        )

    def update_registry_proto(self, registry_proto: RegistryProto):
        self._write_registry(registry_proto)

    def teardown(self):
        self.blob.delete_blob()

    def _write_registry(self, registry_proto: RegistryProto):
        registry_proto.version_id = str(uuid.uuid4())
        registry_proto.last_updated.FromDatetime(_utc_now())

        file_obj = TemporaryFile()
        file_obj.write(registry_proto.SerializeToString())
        file_obj.seek(0)
        self.blob.upload_blob(file_obj, overwrite=True)  # type: ignore
        return

    def set_project_metadata(self, project: str, key: str, value: str):
        registry_proto = self.get_registry_proto()
        found = False
        for pm in registry_proto.project_metadata:
            if pm.project == project:
                try:
                    meta = json.loads(pm.project_uuid) if pm.project_uuid else {}
                except Exception:
                    meta = {}
                if not isinstance(meta, dict):
                    meta = {}
                meta[key] = value
                pm.project_uuid = json.dumps(meta)
                found = True
                break
        if not found:
            from feast.project_metadata import ProjectMetadata

            pm = ProjectMetadata(project_name=project)
            pm.project_uuid = json.dumps({key: value})
            registry_proto.project_metadata.append(pm.to_proto())
        self.update_registry_proto(registry_proto)

    def get_project_metadata(self, project: str, key: str) -> Optional[str]:
        registry_proto = self.get_registry_proto()
        for pm in registry_proto.project_metadata:
            if pm.project == project:
                try:
                    meta = json.loads(pm.project_uuid) if pm.project_uuid else {}
                except Exception:
                    meta = {}
                if not isinstance(meta, dict):
                    return None
                return meta.get(key, None)
        return None



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Copyright 2021 The Feast Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class GEValidationProfiler(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class UserDefinedProfiler(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        BODY_FIELD_NUMBER: builtins.int
        body: builtins.bytes
        """The python-syntax function body (serialized by dill)"""
        def __init__(
            self,
            *,
            body: builtins.bytes = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["body", b"body"]) -> None: ...

    PROFILER_FIELD_NUMBER: builtins.int
    @property
    def profiler(self) -> global___GEValidationProfiler.UserDefinedProfiler: ...
    def __init__(
        self,
        *,
        profiler: global___GEValidationProfiler.UserDefinedProfiler | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["profiler", b"profiler"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["profiler", b"profiler"]) -> None: ...

global___GEValidationProfiler = GEValidationProfiler

class GEValidationProfile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EXPECTATION_SUITE_FIELD_NUMBER: builtins.int
    expectation_suite: builtins.bytes
    """JSON-serialized ExpectationSuite object"""
    def __init__(
        self,
        *,
        expectation_suite: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["expectation_suite", b"expectation_suite"]) -> None: ...

global___GEValidationProfile = GEValidationProfile

class ValidationReference(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class TagsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    REFERENCE_DATASET_NAME_FIELD_NUMBER: builtins.int
    PROJECT_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    TAGS_FIELD_NUMBER: builtins.int
    GE_PROFILER_FIELD_NUMBER: builtins.int
    GE_PROFILE_FIELD_NUMBER: builtins.int
    name: builtins.str
    """Unique name of validation reference within the project"""
    reference_dataset_name: builtins.str
    """Name of saved dataset used as reference dataset"""
    project: builtins.str
    """Name of Feast project that this object source belongs to"""
    description: builtins.str
    """Description of the validation reference"""
    @property
    def tags(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """User defined metadata"""
    @property
    def ge_profiler(self) -> global___GEValidationProfiler: ...
    @property
    def ge_profile(self) -> global___GEValidationProfile: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        reference_dataset_name: builtins.str = ...,
        project: builtins.str = ...,
        description: builtins.str = ...,
        tags: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        ge_profiler: global___GEValidationProfiler | None = ...,
        ge_profile: global___GEValidationProfile | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["cached_profile", b"cached_profile", "ge_profile", b"ge_profile", "ge_profiler", b"ge_profiler", "profiler", b"profiler"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["cached_profile", b"cached_profile", "description", b"description", "ge_profile", b"ge_profile", "ge_profiler", b"ge_profiler", "name", b"name", "profiler", b"profiler", "project", b"project", "reference_dataset_name", b"reference_dataset_name", "tags", b"tags"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["cached_profile", b"cached_profile"]) -> typing_extensions.Literal["ge_profile"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["profiler", b"profiler"]) -> typing_extensions.Literal["ge_profiler"] | None: ...

global___ValidationReference = ValidationReference



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/Store.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16\x66\x65\x61st/core/Store.proto\x12\nfeast.core\"\xfd\x06\n\x05Store\x12\x0c\n\x04name\x18\x01 \x01(\t\x12)\n\x04type\x18\x02 \x01(\x0e\x32\x1b.feast.core.Store.StoreType\x12\x35\n\rsubscriptions\x18\x04 \x03(\x0b\x32\x1e.feast.core.Store.Subscription\x12\x35\n\x0credis_config\x18\x0b \x01(\x0b\x32\x1d.feast.core.Store.RedisConfigH\x00\x12\x44\n\x14redis_cluster_config\x18\x0e \x01(\x0b\x32$.feast.core.Store.RedisClusterConfigH\x00\x1a\x88\x01\n\x0bRedisConfig\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\x12\x1a\n\x12initial_backoff_ms\x18\x03 \x01(\x05\x12\x13\n\x0bmax_retries\x18\x04 \x01(\x05\x12\x1f\n\x17\x66lush_frequency_seconds\x18\x05 \x01(\x05\x12\x0b\n\x03ssl\x18\x06 \x01(\x08\x1a\xdb\x02\n\x12RedisClusterConfig\x12\x19\n\x11\x63onnection_string\x18\x01 \x01(\t\x12\x1a\n\x12initial_backoff_ms\x18\x02 \x01(\x05\x12\x13\n\x0bmax_retries\x18\x03 \x01(\x05\x12\x1f\n\x17\x66lush_frequency_seconds\x18\x04 \x01(\x05\x12\x12\n\nkey_prefix\x18\x05 \x01(\t\x12\x17\n\x0f\x65nable_fallback\x18\x06 \x01(\x08\x12\x17\n\x0f\x66\x61llback_prefix\x18\x07 \x01(\t\x12@\n\tread_from\x18\x08 \x01(\x0e\x32-.feast.core.Store.RedisClusterConfig.ReadFrom\"P\n\x08ReadFrom\x12\n\n\x06MASTER\x10\x00\x12\x14\n\x10MASTER_PREFERRED\x10\x01\x12\x0b\n\x07REPLICA\x10\x02\x12\x15\n\x11REPLICA_PREFERRED\x10\x03\x1a\x44\n\x0cSubscription\x12\x0f\n\x07project\x18\x03 \x01(\t\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07\x65xclude\x18\x04 \x01(\x08J\x04\x08\x02\x10\x03\"N\n\tStoreType\x12\x0b\n\x07INVALID\x10\x00\x12\t\n\x05REDIS\x10\x01\x12\x11\n\rREDIS_CLUSTER\x10\x04\"\x04\x08\x02\x10\x02\"\x04\x08\x03\x10\x03\"\x04\x08\x0c\x10\x0c\"\x04\x08\r\x10\rB\x08\n\x06\x63onfigBO\n\x10\x66\x65\x61st.proto.coreB\nStoreProtoZ/github.com/feast-dev/feast/go/protos/feast/coreb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.core.Store_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020feast.proto.coreB\nStoreProtoZ/github.com/feast-dev/feast/go/protos/feast/core'
  _globals['_STORE']._serialized_start=39
  _globals['_STORE']._serialized_end=932
  _globals['_STORE_REDISCONFIG']._serialized_start=286
  _globals['_STORE_REDISCONFIG']._serialized_end=422
  _globals['_STORE_REDISCLUSTERCONFIG']._serialized_start=425
  _globals['_STORE_REDISCLUSTERCONFIG']._serialized_end=772
  _globals['_STORE_REDISCLUSTERCONFIG_READFROM']._serialized_start=692
  _globals['_STORE_REDISCLUSTERCONFIG_READFROM']._serialized_end=772
  _globals['_STORE_SUBSCRIPTION']._serialized_start=774
  _globals['_STORE_SUBSCRIPTION']._serialized_end=842
  _globals['_STORE_STORETYPE']._serialized_start=844
  _globals['_STORE_STORETYPE']._serialized_end=922
# @@protoc_insertion_point(module_scope)



from pydantic import BaseModel, Field


class SupportedAsyncMethods(BaseModel):
    read: bool = Field(default=False)
    write: bool = Field(default=False)


class ProviderAsyncMethods(BaseModel):
    online: SupportedAsyncMethods = Field(default_factory=SupportedAsyncMethods)



"""
A module with utility functions and classes to support authorizing the Arrow Flight servers.
"""

import asyncio
import functools
import logging
from typing import cast

import pyarrow.flight as fl
from pyarrow.flight import ServerCallContext

from feast.permissions.auth.auth_manager import (
    get_auth_manager,
)
from feast.permissions.security_manager import get_security_manager
from feast.permissions.user import User

logger = logging.getLogger(__name__)


class AuthorizationMiddlewareFactory(fl.ServerMiddlewareFactory):
    """
    A middleware factory to intercept the authorization header and propagate it to the authorization middleware.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def start_call(self, info, headers):
        """
        Intercept the authorization header and propagate it to the authorization middleware.
        """
        access_token = get_auth_manager().token_extractor.extract_access_token(
            headers=headers
        )
        return AuthorizationMiddleware(access_token=access_token)


class AuthorizationMiddleware(fl.ServerMiddleware):
    """
    A server middleware holding the authorization header and offering a method to extract the user credentials.
    """

    def __init__(self, access_token: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.access_token = access_token

    def call_completed(self, exception):
        if exception:
            logger.exception(
                f"{AuthorizationMiddleware.__name__} encountered an exception: {exception}"
            )

    async def extract_user(self) -> User:
        """
        Use the configured `TokenParser` to extract the user credentials.
        """
        return await get_auth_manager().token_parser.user_details_from_access_token(
            self.access_token
        )


def inject_user_details(context: ServerCallContext):
    """
    Function to use in Arrow Flight endpoints (e.g. `do_get`, `do_put` and so on) to access the token extracted from the header,
    extract the user details out of it and propagate them to the current security manager, if any.

    Args:
        context: The endpoint context.
    """
    if context.get_middleware("auth") is None:
        logger.warning("No `auth` middleware.")
        return

    sm = get_security_manager()
    if sm is not None:
        auth_middleware = cast(AuthorizationMiddleware, context.get_middleware("auth"))
        current_user = asyncio.run(auth_middleware.extract_user())
        logger.debug(f"User extracted: {current_user}")

        sm.set_current_user(current_user)


def inject_user_details_decorator(func):
    @functools.wraps(func)
    def wrapper(self, context, *args, **kwargs):
        inject_user_details(context)
        return func(self, context, *args, **kwargs)

    return wrapper



import json
import logging
import re

from fastapi import Depends, FastAPI, HTTPException, Request, status
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from pydantic import ValidationError

from feast import FeatureStore
from feast.api.registry.rest import register_all_routes
from feast.errors import (
    FeastObjectNotFoundException,
    FeastPermissionError,
    PushSourceNotFoundException,
)
from feast.permissions.auth.auth_manager import get_auth_manager
from feast.permissions.server.rest import inject_user_details
from feast.permissions.server.utils import (
    ServerType,
    init_auth_manager,
    init_security_manager,
    str_to_auth_manager_type,
)
from feast.registry_server import RegistryServer
from feast.utils import _utc_now

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class RestRegistryServer:
    def __init__(self, store: FeatureStore):
        self.store = store
        self.registry = store.registry
        self.grpc_handler = RegistryServer(self.registry)
        recent_visit_logging_cfg = {}
        feature_server_cfg = getattr(
            getattr(store, "config", None), "feature_server", None
        )
        if isinstance(feature_server_cfg, dict):
            recent_visit_logging_cfg = feature_server_cfg.get(
                "recent_visit_logging", {}
            )
        self.recent_visits_limit = recent_visit_logging_cfg.get("limit", 100)
        self.log_patterns = recent_visit_logging_cfg.get(
            "log_patterns",
            [
                r".*/entities/(?!all$)[^/]+$",
                r".*/data_sources/(?!all$)[^/]+$",
                r".*/feature_views/(?!all$)[^/]+$",
                r".*/features/(?!all$)[^/]+$",
                r".*/feature_services/(?!all$)[^/]+$",
                r".*/saved_datasets/(?!all$)[^/]+$",
            ],
        )
        self.app = FastAPI(
            title="Feast REST Registry Server",
            description="Feast REST Registry Server",
            dependencies=[Depends(inject_user_details)],
            version="1.0.0",
            openapi_url="/openapi.json",
            root_path="/api/v1",
            docs_url="/",
            redoc_url="/docs",
            default_status_code=status.HTTP_200_OK,
            default_headers={
                "X-Content-Type-Options": "nosniff",
                "X-XSS-Protection": "1; mode=block",
                "X-Frame-Options": "DENY",
            },
        )
        self._add_exception_handlers()
        self._add_logging_middleware()
        self._add_openapi_security()
        self._init_auth()
        self._register_routes()

    def _add_exception_handlers(self):
        """Add custom exception handlers to include HTTP status codes in JSON responses."""

        @self.app.exception_handler(HTTPException)
        async def http_exception_handler(request: Request, exc: HTTPException):
            return JSONResponse(
                status_code=exc.status_code,
                content={
                    "status_code": exc.status_code,
                    "detail": exc.detail,
                    "error_type": "HTTPException",
                },
            )

        @self.app.exception_handler(FeastObjectNotFoundException)
        async def feast_object_not_found_handler(
            request: Request, exc: FeastObjectNotFoundException
        ):
            return JSONResponse(
                status_code=404,
                content={
                    "status_code": 404,
                    "detail": str(exc),
                    "error_type": "FeastObjectNotFoundException",
                },
            )

        @self.app.exception_handler(FeastPermissionError)
        async def feast_permission_error_handler(
            request: Request, exc: FeastPermissionError
        ):
            return JSONResponse(
                status_code=403,
                content={
                    "status_code": 403,
                    "detail": str(exc),
                    "error_type": "FeastPermissionError",
                },
            )

        @self.app.exception_handler(PushSourceNotFoundException)
        async def push_source_not_found_handler(
            request: Request, exc: PushSourceNotFoundException
        ):
            return JSONResponse(
                status_code=422,
                content={
                    "status_code": 422,
                    "detail": str(exc),
                    "error_type": "PushSourceNotFoundException",
                },
            )

        @self.app.exception_handler(ValidationError)
        async def validation_error_handler(request: Request, exc: ValidationError):
            return JSONResponse(
                status_code=422,
                content={
                    "status_code": 422,
                    "detail": str(exc),
                    "error_type": "ValidationError",
                },
            )

        @self.app.exception_handler(RequestValidationError)
        async def request_validation_error_handler(
            request: Request, exc: RequestValidationError
        ):
            return JSONResponse(
                status_code=422,
                content={
                    "status_code": 422,
                    "detail": str(exc),
                    "error_type": "RequestValidationError",
                },
            )

        @self.app.exception_handler(ValueError)
        async def value_error_handler(request: Request, exc: ValueError):
            return JSONResponse(
                status_code=422,
                content={
                    "status_code": 422,
                    "detail": str(exc),
                    "error_type": "ValueError",
                },
            )

        @self.app.exception_handler(Exception)
        async def generic_exception_handler(request: Request, exc: Exception):
            logger.error(f"Unhandled exception: {exc}", exc_info=True)

            return JSONResponse(
                status_code=500,
                content={
                    "status_code": 500,
                    "detail": f"Internal server error: {str(exc)}",
                    "error_type": "InternalServerError",
                },
            )

    def _add_logging_middleware(self):
        from fastapi import Request
        from starlette.middleware.base import BaseHTTPMiddleware

        class LoggingMiddleware(BaseHTTPMiddleware):
            def __init__(
                self, app, registry, project, recent_visits_limit, log_patterns
            ):
                super().__init__(app)
                self.registry = registry
                self.project = project
                self.recent_visits_limit = recent_visits_limit
                self.log_patterns = [re.compile(p) for p in log_patterns]

            async def dispatch(self, request: Request, call_next):
                LOG_PATTERNS = self.log_patterns
                if request.method == "GET":
                    user = None
                    if hasattr(request.state, "user"):
                        user = getattr(request.state, "user", None)
                    if not user:
                        user = "anonymous"
                    project = request.query_params.get("project") or self.project
                    key = f"recently_visited_{user}"
                    path = str(request.url.path)
                    method = request.method

                    if path.startswith("/api/v1/metrics/"):
                        response = await call_next(request)
                        return response

                    object_type = None
                    object_name = None
                    if not any(p.match(path) for p in LOG_PATTERNS):
                        response = await call_next(request)
                        return response
                    m = re.match(r"/api/v1/([^/]+)(?:/([^/]+))?", path)
                    if m:
                        object_type = m.group(1)
                        object_name = m.group(2)
                    else:
                        object_type = None
                        object_name = None
                    visit = {
                        "path": path,
                        "timestamp": _utc_now().isoformat(),
                        "project": project,
                        "user": user,
                        "object": object_type,
                        "object_name": object_name,
                        "method": method,
                    }
                    try:
                        visits_json = self.registry.get_project_metadata(project, key)
                        visits = json.loads(visits_json) if visits_json else []
                    except Exception:
                        visits = []
                    visits.append(visit)
                    visits = visits[-self.recent_visits_limit :]
                    try:
                        self.registry.set_project_metadata(
                            project, key, json.dumps(visits)
                        )
                    except Exception as e:
                        logger.warning(f"Failed to persist recent visits: {e}")
                response = await call_next(request)
                return response

        self.app.add_middleware(
            LoggingMiddleware,
            registry=self.registry,
            project=self.store.project,
            recent_visits_limit=self.recent_visits_limit,
            log_patterns=self.log_patterns,
        )

    def _register_routes(self):
        register_all_routes(self.app, self.grpc_handler, self)

    def _add_openapi_security(self):
        if self.app.openapi_schema:
            return
        original_openapi = self.app.openapi

        def custom_openapi():
            if self.app.openapi_schema:
                return self.app.openapi_schema
            schema = original_openapi()
            schema.setdefault("components", {}).setdefault("securitySchemes", {})[
                "BearerAuth"
            ] = {
                "type": "http",
                "scheme": "bearer",
                "bearerFormat": "JWT",
            }
            schema.setdefault("security", []).append({"BearerAuth": []})
            self.app.openapi_schema = schema
            return self.app.openapi_schema

        self.app.openapi = custom_openapi

    def _init_auth(self):
        auth_type = str_to_auth_manager_type(self.store.config.auth_config.type)
        init_security_manager(auth_type=auth_type, fs=self.store)
        init_auth_manager(
            auth_type=auth_type,
            server_type=ServerType.REST,
            auth_config=self.store.config.auth_config,
        )
        self.auth_manager = get_auth_manager()

    def start_server(
        self,
        port: int,
        tls_key_path: str = "",
        tls_cert_path: str = "",
    ):
        import uvicorn

        if tls_key_path and tls_cert_path:
            logger.info("Starting REST registry server in TLS(SSL) mode")
            logger.info(f"REST registry server listening on https://localhost:{port}")
            uvicorn.run(
                self.app,
                host="0.0.0.0",
                port=port,
                ssl_keyfile=tls_key_path,
                ssl_certfile=tls_cert_path,
            )
        else:
            logger.info("Starting REST registry server in non-TLS(SSL) mode")
            logger.info(f"REST registry server listening on http://localhost:{port}")
            uvicorn.run(
                self.app,
                host="0.0.0.0",
                port=port,
            )



from datetime import datetime
from typing import Dict

import numpy as np
import pandas as pd


def create_document_chunks_df(
    documents: Dict[str, str],
    start_date: datetime,
    end_date: datetime,
    embedding_size: int = 60,
) -> pd.DataFrame:
    """
    Example df generated by this function:

    | event_timestamp  | document_id | chunk_id | chunk_text       | embedding | created          |
    |------------------+-------------+----------+------------------+-----------+------------------|
    | 2021-03-17 19:31 | doc_1       | chunk-1  | Hello world      | [0.1, ...]| 2021-03-24 19:34 |
    | 2021-03-17 19:31 | doc_1       | chunk-2  | How are you?     | [0.2, ...]| 2021-03-24 19:34 |
    | 2021-03-17 19:31 | doc_2       | chunk-1  | This is a test   | [0.3, ...]| 2021-03-24 19:34 |
    | 2021-03-17 19:31 | doc_2       | chunk-2  | Document chunk   | [0.4, ...]| 2021-03-24 19:34 |
    """
    df_hourly = pd.DataFrame(
        {
            "event_timestamp": [
                pd.Timestamp(dt, unit="ms").round("ms")
                for dt in pd.date_range(
                    start=start_date,
                    end=end_date,
                    freq="1h",
                    inclusive="left",
                    tz="UTC",
                )
            ]
            + [
                pd.Timestamp(
                    year=2021, month=4, day=12, hour=7, minute=0, second=0, tz="UTC"
                )
            ]
        }
    )
    df_all_chunks = pd.DataFrame()

    for doc_id, doc_text in documents.items():
        chunks = doc_text.split(". ")  # Simple chunking by sentence
        for chunk_id, chunk_text in enumerate(chunks, start=1):
            df_hourly_copy = df_hourly.copy()
            df_hourly_copy["document_id"] = doc_id
            df_hourly_copy["chunk_id"] = f"chunk-{chunk_id}"
            df_hourly_copy["chunk_text"] = chunk_text
            df_all_chunks = pd.concat([df_hourly_copy, df_all_chunks])

    df_all_chunks.reset_index(drop=True, inplace=True)
    rows = df_all_chunks["event_timestamp"].count()

    # Generate random embeddings for each chunk
    df_all_chunks["embedding"] = [
        np.random.rand(embedding_size).tolist() for _ in range(rows)
    ]
    df_all_chunks["created"] = pd.to_datetime(pd.Timestamp.now(tz=None).round("ms"))

    # Create duplicate rows that should be filtered by created timestamp
    late_row = df_all_chunks[rows // 2 : rows // 2 + 1]
    df_all_chunks = pd.concat([df_all_chunks, late_row, late_row], ignore_index=True)

    return df_all_chunks



from feast.importer import import_class

TRANSFORMATION_CLASS_FOR_TYPE = {
    "python": "feast.transformation.python_transformation.PythonTransformation",
    "pandas": "feast.transformation.pandas_transformation.PandasTransformation",
    "substrait": "feast.transformation.substrait_transformation.SubstraitTransformation",
    "sql": "feast.transformation.sql_transformation.SQLTransformation",
    "spark_sql": "feast.transformation.spark_transformation.SparkTransformation",
    "spark": "feast.transformation.spark_transformation.SparkTransformation",
}


def get_transformation_class_from_type(transformation_type: str):
    if transformation_type in TRANSFORMATION_CLASS_FOR_TYPE:
        transformation_type = TRANSFORMATION_CLASS_FOR_TYPE[transformation_type]
    elif not transformation_type.endswith("Transformation"):
        raise ValueError(
            f"Invalid transformation type: {transformation_type}. Choose from {list(TRANSFORMATION_CLASS_FOR_TYPE.keys())}."
        )
    module_name, transformation_class_type = transformation_type.rsplit(".", 1)
    return import_class(
        module_name, transformation_class_type, transformation_class_type
    )



# Copyright 2021 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Callable, Dict, List, Mapping, Optional, Sequence, Tuple, Union

from feast import Entity, utils
from feast.batch_feature_view import BatchFeatureView
from feast.feature_service import FeatureService
from feast.feature_view import FeatureView
from feast.infra.infra_object import InfraObject
from feast.infra.registry.base_registry import BaseRegistry
from feast.infra.supported_async_methods import SupportedAsyncMethods
from feast.online_response import OnlineResponse
from feast.protos.feast.core.Registry_pb2 import Registry as RegistryProto
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import RepeatedValue
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import RepoConfig
from feast.stream_feature_view import StreamFeatureView


class OnlineStore(ABC):
    """
    The interface that Feast uses to interact with the storage system that handles online features.
    """

    @property
    def async_supported(self) -> SupportedAsyncMethods:
        return SupportedAsyncMethods()

    @abstractmethod
    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        """
        Writes a batch of feature rows to the online store.

        If a tz-naive timestamp is passed to this method, it is assumed to be UTC.

        Args:
            config: The config for the current feature store.
            table: Feature view to which these feature rows correspond.
            data: A list of quadruplets containing feature data. Each quadruplet contains an entity
                key, a dict containing feature values, an event timestamp for the row, and the created
                timestamp for the row if it exists.
            progress: Function to be called once a batch of rows is written to the online store, used
                to show progress.
        """
        pass

    async def online_write_batch_async(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        """
        Writes a batch of feature rows to the online store asynchronously.

        If a tz-naive timestamp is passed to this method, it is assumed to be UTC.

        Args:
            config: The config for the current feature store.
            table: Feature view to which these feature rows correspond.
            data: A list of quadruplets containing feature data. Each quadruplet contains an entity
                key, a dict containing feature values, an event timestamp for the row, and the created
                timestamp for the row if it exists.
            progress: Function to be called once a batch of rows is written to the online store, used
                to show progress.
        """
        raise NotImplementedError(
            f"Online store {self.__class__.__name__} does not support online write batch async"
        )

    @abstractmethod
    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        """
        Reads features values for the given entity keys.

        Args:
            config: The config for the current feature store.
            table: The feature view whose feature values should be read.
            entity_keys: The list of entity keys for which feature values should be read.
            requested_features: The list of features that should be read.

        Returns:
            A list of the same length as entity_keys. Each item in the list is a tuple where the first
            item is the event timestamp for the row, and the second item is a dict mapping feature names
            to values, which are returned in proto format.
        """
        pass

    async def online_read_async(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        """
        Reads features values for the given entity keys asynchronously.

        Args:
            config: The config for the current feature store.
            table: The feature view whose feature values should be read.
            entity_keys: The list of entity keys for which feature values should be read.
            requested_features: The list of features that should be read.

        Returns:
            A list of the same length as entity_keys. Each item in the list is a tuple where the first
            item is the event timestamp for the row, and the second item is a dict mapping feature names
            to values, which are returned in proto format.
        """
        raise NotImplementedError(
            f"Online store {self.__class__.__name__} does not support online read async"
        )

    def get_online_features(
        self,
        config: RepoConfig,
        features: Union[List[str], FeatureService],
        entity_rows: Union[
            List[Dict[str, Any]],
            Mapping[str, Union[Sequence[Any], Sequence[ValueProto], RepeatedValue]],
        ],
        registry: BaseRegistry,
        project: str,
        full_feature_names: bool = False,
    ) -> OnlineResponse:
        if isinstance(entity_rows, list):
            columnar: Dict[str, List[Any]] = {k: [] for k in entity_rows[0].keys()}
            for entity_row in entity_rows:
                for key, value in entity_row.items():
                    try:
                        columnar[key].append(value)
                    except KeyError as e:
                        raise ValueError(
                            "All entity_rows must have the same keys."
                        ) from e

            entity_rows = columnar

        (
            join_key_values,
            grouped_refs,
            entity_name_to_join_key_map,
            requested_on_demand_feature_views,
            feature_refs,
            requested_result_row_names,
            online_features_response,
        ) = utils._prepare_entities_to_read_from_online_store(
            registry=registry,
            project=project,
            features=features,
            entity_values=entity_rows,
            full_feature_names=full_feature_names,
            native_entity_values=True,
        )

        for table, requested_features in grouped_refs:
            # Get the correct set of entity values with the correct join keys.
            table_entity_values, idxs, output_len = utils._get_unique_entities(
                table,
                join_key_values,
                entity_name_to_join_key_map,
            )

            entity_key_protos = utils._get_entity_key_protos(table_entity_values)

            # Fetch data for Entities.
            read_rows = self.online_read(
                config=config,
                table=table,
                entity_keys=entity_key_protos,
                requested_features=requested_features,
            )

            feature_data = utils._convert_rows_to_protobuf(
                requested_features, read_rows
            )

            # Populate the result_rows with the Features from the OnlineStore inplace.
            utils._populate_response_from_feature_data(
                feature_data,
                idxs,
                online_features_response,
                full_feature_names,
                requested_features,
                table,
                output_len,
            )

        if requested_on_demand_feature_views:
            utils._augment_response_with_on_demand_transforms(
                online_features_response,
                feature_refs,
                requested_on_demand_feature_views,
                full_feature_names,
            )

        utils._drop_unneeded_columns(
            online_features_response, requested_result_row_names
        )
        return OnlineResponse(online_features_response)

    async def get_online_features_async(
        self,
        config: RepoConfig,
        features: Union[List[str], FeatureService],
        entity_rows: Union[
            List[Dict[str, Any]],
            Mapping[str, Union[Sequence[Any], Sequence[ValueProto], RepeatedValue]],
        ],
        registry: BaseRegistry,
        project: str,
        full_feature_names: bool = False,
    ) -> OnlineResponse:
        if isinstance(entity_rows, list):
            columnar: Dict[str, List[Any]] = {k: [] for k in entity_rows[0].keys()}
            for entity_row in entity_rows:
                for key, value in entity_row.items():
                    try:
                        columnar[key].append(value)
                    except KeyError as e:
                        raise ValueError(
                            "All entity_rows must have the same keys."
                        ) from e

            entity_rows = columnar

        (
            join_key_values,
            grouped_refs,
            entity_name_to_join_key_map,
            requested_on_demand_feature_views,
            feature_refs,
            requested_result_row_names,
            online_features_response,
        ) = utils._prepare_entities_to_read_from_online_store(
            registry=registry,
            project=project,
            features=features,
            entity_values=entity_rows,
            full_feature_names=full_feature_names,
            native_entity_values=True,
        )

        async def query_table(table, requested_features):
            # Get the correct set of entity values with the correct join keys.
            table_entity_values, idxs, output_len = utils._get_unique_entities(
                table,
                join_key_values,
                entity_name_to_join_key_map,
            )

            entity_key_protos = utils._get_entity_key_protos(table_entity_values)

            # Fetch data for Entities.
            read_rows = await self.online_read_async(
                config=config,
                table=table,
                entity_keys=entity_key_protos,
                requested_features=requested_features,
            )

            return idxs, read_rows, output_len

        all_responses = await asyncio.gather(
            *[
                query_table(table, requested_features)
                for table, requested_features in grouped_refs
            ]
        )

        for (idxs, read_rows, output_len), (table, requested_features) in zip(
            all_responses, grouped_refs
        ):
            feature_data = utils._convert_rows_to_protobuf(
                requested_features, read_rows
            )

            # Populate the result_rows with the Features from the OnlineStore inplace.
            utils._populate_response_from_feature_data(
                feature_data,
                idxs,
                online_features_response,
                full_feature_names,
                requested_features,
                table,
                output_len,
            )

        if requested_on_demand_feature_views:
            utils._augment_response_with_on_demand_transforms(
                online_features_response,
                feature_refs,
                requested_on_demand_feature_views,
                full_feature_names,
            )

        utils._drop_unneeded_columns(
            online_features_response, requested_result_row_names
        )
        return OnlineResponse(online_features_response)

    @abstractmethod
    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[
            Union[BatchFeatureView, StreamFeatureView, FeatureView]
        ],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        """
        Reconciles cloud resources with the specified set of Feast objects.

        Args:
            config: The config for the current feature store.
            tables_to_delete: Feature views whose corresponding infrastructure should be deleted.
            tables_to_keep: Feature views whose corresponding infrastructure should not be deleted, and
                may need to be updated.
            entities_to_delete: Entities whose corresponding infrastructure should be deleted.
            entities_to_keep: Entities whose corresponding infrastructure should not be deleted, and
                may need to be updated.
            partial: If true, tables_to_delete and tables_to_keep are not exhaustive lists, so
                infrastructure corresponding to other feature views should be not be touched.
        """
        pass

    def plan(
        self, config: RepoConfig, desired_registry_proto: RegistryProto
    ) -> List[InfraObject]:
        """
        Returns the set of InfraObjects required to support the desired registry.

        Args:
            config: The config for the current feature store.
            desired_registry_proto: The desired registry, in proto form.
        """
        return []

    @abstractmethod
    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ):
        """
        Tears down all cloud resources for the specified set of Feast objects.

        Args:
            config: The config for the current feature store.
            tables: Feature views whose corresponding infrastructure should be deleted.
            entities: Entities whose corresponding infrastructure should be deleted.
        """
        pass

    def retrieve_online_documents(
        self,
        config: RepoConfig,
        table: FeatureView,
        requested_features: List[str],
        embedding: List[float],
        top_k: int,
        distance_metric: Optional[str] = None,
    ) -> List[
        Tuple[
            Optional[datetime],
            Optional[EntityKeyProto],
            Optional[ValueProto],
            Optional[ValueProto],
            Optional[ValueProto],
        ]
    ]:
        """
        Retrieves online feature values for the specified embeddings.

        Args:
            distance_metric: distance metric to use for retrieval.
            config: The config for the current feature store.
            table: The feature view whose feature values should be read.
            requested_features: The list of features whose embeddings should be used for retrieval.
            embedding: The embeddings to use for retrieval.
            top_k: The number of documents to retrieve.

        Returns:
            object: A list of top k closest documents to the specified embedding. Each item in the list is a tuple
            where the first item is the event timestamp for the row, and the second item is a dict of feature
            name to embeddings.
        """
        if not requested_features:
            raise ValueError("Requested_features must be specified")
        raise NotImplementedError(
            f"Online store {self.__class__.__name__} does not support online retrieval"
        )

    def retrieve_online_documents_v2(
        self,
        config: RepoConfig,
        table: FeatureView,
        requested_features: List[str],
        embedding: Optional[List[float]],
        top_k: int,
        distance_metric: Optional[str] = None,
        query_string: Optional[str] = None,
    ) -> List[
        Tuple[
            Optional[datetime],
            Optional[EntityKeyProto],
            Optional[Dict[str, ValueProto]],
        ]
    ]:
        """
        Retrieves online feature values for the specified embeddings.

        Args:
            distance_metric: distance metric to use for retrieval.
            config: The config for the current feature store.
            table: The feature view whose feature values should be read.
            requested_features: The list of features whose embeddings should be used for retrieval.
            embedding: The embeddings to use for retrieval (optional)
            top_k: The number of documents to retrieve.
            query_string: The query string to search for using keyword search (bm25) (optional)

        Returns:
            object: A list of top k closest documents to the specified embedding. Each item in the list is a tuple
            where the first item is the event timestamp for the row, and the second item is a dict of feature
            name to embeddings.
        """
        assert embedding is not None or query_string is not None, (
            "Either embedding or query_string must be specified"
        )
        raise NotImplementedError(
            f"Online store {self.__class__.__name__} does not support online retrieval"
        )

    async def initialize(self, config: RepoConfig) -> None:
        pass

    async def close(self) -> None:
        pass



from fastapi.requests import Request
from starlette.authentication import (
    AuthenticationError,
)

from feast.permissions.auth.token_extractor import TokenExtractor


class RestTokenExtractor(TokenExtractor):
    def extract_access_token(self, **kwargs) -> str:
        """
        Token extractor for REST requests.

        Requires a keyword argument called `request` of type `Request`

        Returns:
            The extracted access token.
        """

        if "request" not in kwargs:
            raise ValueError("Missing keywork argument 'request'")
        if not isinstance(kwargs["request"], Request):
            raise ValueError(
                f"The keywork argument 'request' is not of the expected type {Request.__name__}"
            )

        access_token = None
        request = kwargs["request"]
        if isinstance(request, Request):
            headers = request.headers
            for header in headers:
                if header.lower() == "authorization":
                    return self._extract_bearer_token(headers[header])

        if access_token is None:
            raise AuthenticationError("Missing authorization header")

        return access_token



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/SavedDataset.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from feast.protos.feast.core import DataSource_pb2 as feast_dot_core_dot_DataSource__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1d\x66\x65\x61st/core/SavedDataset.proto\x12\nfeast.core\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1b\x66\x65\x61st/core/DataSource.proto\"\xa5\x02\n\x10SavedDatasetSpec\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07project\x18\x02 \x01(\t\x12\x10\n\x08\x66\x65\x61tures\x18\x03 \x03(\t\x12\x11\n\tjoin_keys\x18\x04 \x03(\t\x12\x1a\n\x12\x66ull_feature_names\x18\x05 \x01(\x08\x12\x30\n\x07storage\x18\x06 \x01(\x0b\x32\x1f.feast.core.SavedDatasetStorage\x12\x1c\n\x14\x66\x65\x61ture_service_name\x18\x08 \x01(\t\x12\x34\n\x04tags\x18\x07 \x03(\x0b\x32&.feast.core.SavedDatasetSpec.TagsEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa9\x04\n\x13SavedDatasetStorage\x12:\n\x0c\x66ile_storage\x18\x04 \x01(\x0b\x32\".feast.core.DataSource.FileOptionsH\x00\x12\x42\n\x10\x62igquery_storage\x18\x05 \x01(\x0b\x32&.feast.core.DataSource.BigQueryOptionsH\x00\x12\x42\n\x10redshift_storage\x18\x06 \x01(\x0b\x32&.feast.core.DataSource.RedshiftOptionsH\x00\x12\x44\n\x11snowflake_storage\x18\x07 \x01(\x0b\x32\'.feast.core.DataSource.SnowflakeOptionsH\x00\x12<\n\rtrino_storage\x18\x08 \x01(\x0b\x32#.feast.core.DataSource.TrinoOptionsH\x00\x12<\n\rspark_storage\x18\t \x01(\x0b\x32#.feast.core.DataSource.SparkOptionsH\x00\x12\x44\n\x0e\x63ustom_storage\x18\n \x01(\x0b\x32*.feast.core.DataSource.CustomSourceOptionsH\x00\x12>\n\x0e\x61thena_storage\x18\x0b \x01(\x0b\x32$.feast.core.DataSource.AthenaOptionsH\x00\x42\x06\n\x04kind\"\xf7\x01\n\x10SavedDatasetMeta\x12\x35\n\x11\x63reated_timestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12:\n\x16last_updated_timestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x37\n\x13min_event_timestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x37\n\x13max_event_timestamp\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"f\n\x0cSavedDataset\x12*\n\x04spec\x18\x01 \x01(\x0b\x32\x1c.feast.core.SavedDatasetSpec\x12*\n\x04meta\x18\x02 \x01(\x0b\x32\x1c.feast.core.SavedDatasetMetaBV\n\x10\x66\x65\x61st.proto.coreB\x11SavedDatasetProtoZ/github.com/feast-dev/feast/go/protos/feast/coreb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.core.SavedDataset_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020feast.proto.coreB\021SavedDatasetProtoZ/github.com/feast-dev/feast/go/protos/feast/core'
  _globals['_SAVEDDATASETSPEC_TAGSENTRY']._options = None
  _globals['_SAVEDDATASETSPEC_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_SAVEDDATASETSPEC']._serialized_start=108
  _globals['_SAVEDDATASETSPEC']._serialized_end=401
  _globals['_SAVEDDATASETSPEC_TAGSENTRY']._serialized_start=358
  _globals['_SAVEDDATASETSPEC_TAGSENTRY']._serialized_end=401
  _globals['_SAVEDDATASETSTORAGE']._serialized_start=404
  _globals['_SAVEDDATASETSTORAGE']._serialized_end=957
  _globals['_SAVEDDATASETMETA']._serialized_start=960
  _globals['_SAVEDDATASETMETA']._serialized_end=1207
  _globals['_SAVEDDATASET']._serialized_start=1209
  _globals['_SAVEDDATASET']._serialized_end=1311
# @@protoc_insertion_point(module_scope)



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime, timezone
from typing import Dict, Optional

from google.protobuf.json_format import MessageToJson
from typeguard import typechecked

from feast.protos.feast.core.Project_pb2 import Project as ProjectProto
from feast.protos.feast.core.Project_pb2 import ProjectMeta as ProjectMetaProto
from feast.protos.feast.core.Project_pb2 import ProjectSpec as ProjectSpecProto
from feast.utils import _utc_now


@typechecked
class Project:
    """
    Project is a collection of Feast Objects. Projects provide complete isolation of
    feature stores at the infrastructure level.

    Attributes:
        name: The unique name of the project.
        description: A human-readable description.
        tags: A dictionary of key-value pairs to store arbitrary metadata.
        owner: The owner of the project, typically the email of the primary maintainer.
        created_timestamp: The time when the entity was created.
        last_updated_timestamp: The time when the entity was last updated.
    """

    name: str
    description: str
    tags: Dict[str, str]
    owner: str
    created_timestamp: datetime
    last_updated_timestamp: datetime

    def __init__(
        self,
        *,
        name: str,
        description: str = "",
        tags: Optional[Dict[str, str]] = None,
        owner: str = "",
        created_timestamp: Optional[datetime] = None,
        last_updated_timestamp: Optional[datetime] = None,
    ):
        """
        Creates Project object.

        Args:
            name: The unique name of the project.
            description (optional): A human-readable description.
            tags (optional): A dictionary of key-value pairs to store arbitrary metadata.
            owner (optional): The owner of the project, typically the email of the primary maintainer.
            created_timestamp (optional): The time when the project was created. Defaults to
            last_updated_timestamp (optional): The time when the project was last updated.

        Raises:
            ValueError: Parameters are specified incorrectly.
        """
        self.name = name
        self.description = description
        self.tags = tags if tags is not None else {}
        self.owner = owner
        updated_time = _utc_now()
        self.created_timestamp = created_timestamp or updated_time
        self.last_updated_timestamp = last_updated_timestamp or updated_time

    def __hash__(self) -> int:
        return hash((self.name))

    def __eq__(self, other):
        if not isinstance(other, Project):
            raise TypeError("Comparisons should only involve Project class objects.")

        if (
            self.name != other.name
            or self.description != other.description
            or self.tags != other.tags
            or self.owner != other.owner
            or self.created_timestamp != other.created_timestamp
            or self.last_updated_timestamp != other.last_updated_timestamp
        ):
            return False

        return True

    def __str__(self):
        return str(MessageToJson(self.to_proto()))

    def __lt__(self, other):
        return self.name < other.name

    def is_valid(self):
        """
        Validates the state of this project locally.

        Raises:
            ValueError: The project does not have a name or does not have a type.
        """
        if not self.name:
            raise ValueError("The project does not have a name.")

        from feast.repo_operations import is_valid_name

        if not is_valid_name(self.name):
            raise ValueError(
                f"Project name, {self.name}, should only have "
                f"alphanumerical values, underscores, and hyphens but not start with an underscore or hyphen."
            )

    @classmethod
    def from_proto(cls, project_proto: ProjectProto):
        """
        Creates a project from a protobuf representation of an project.

        Args:
            entity_proto: A protobuf representation of an project.

        Returns:
            An Entity object based on the entity protobuf.
        """
        project = cls(
            name=project_proto.spec.name,
            description=project_proto.spec.description,
            tags=dict(project_proto.spec.tags),
            owner=project_proto.spec.owner,
        )
        if project_proto.meta.HasField("created_timestamp"):
            project.created_timestamp = (
                project_proto.meta.created_timestamp.ToDatetime().replace(
                    tzinfo=timezone.utc
                )
            )
        if project_proto.meta.HasField("last_updated_timestamp"):
            project.last_updated_timestamp = (
                project_proto.meta.last_updated_timestamp.ToDatetime().replace(
                    tzinfo=timezone.utc
                )
            )

        return project

    def to_proto(self) -> ProjectProto:
        """
        Converts an project object to its protobuf representation.

        Returns:
            An ProjectProto protobuf.
        """
        meta = ProjectMetaProto()
        if self.created_timestamp:
            meta.created_timestamp.FromDatetime(self.created_timestamp)
        if self.last_updated_timestamp:
            meta.last_updated_timestamp.FromDatetime(self.last_updated_timestamp)

        spec = ProjectSpecProto(
            name=self.name,
            description=self.description,
            tags=self.tags,
            owner=self.owner,
        )

        return ProjectProto(spec=spec, meta=meta)



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Optional

import numpy as np
from transformers import (
    PreTrainedModel,
    PreTrainedTokenizer,
    RagRetriever,
)

from feast import FeatureStore, FeatureView
from feast.torch_wrapper import get_torch
from feast.vector_store import FeastVectorStore


class FeastIndex:
    """Dummy index required by HuggingFace's RagRetriever."""

    def __init__(self):
        """Initialize the Feast index."""
        pass


class FeastRAGRetriever(RagRetriever):
    """RAG retriever implementation that uses Feast as a backend."""

    VALID_SEARCH_TYPES = {"text", "vector", "hybrid"}

    def __init__(
        self,
        question_encoder_tokenizer: PreTrainedTokenizer,
        question_encoder: PreTrainedModel,
        generator_tokenizer: PreTrainedTokenizer,
        generator_model: PreTrainedModel,
        feast_repo_path: str,
        feature_view: FeatureView,
        features: list[str],
        search_type: str,
        config: dict[str, Any],
        index: FeastIndex,
        format_document: Optional[Callable[[dict[str, Any]], str]] = None,
        id_field: str = "",
        text_field: str = "",
        **kwargs,
    ):
        """Initialize the Feast RAG retriever.

        Args:
            question_encoder_tokenizer: Tokenizer for encoding questions
            question_encoder: Model for encoding questions
            generator_tokenizer: Tokenizer for the generator model
            generator_model: The generator model
            feast_repo_path: Path to the Feast repository
            feature_view: Feast FeatureView containing the document data
            features: List of feature names to use from the feature view
            search_type: Type of search to perform (text, vector, or hybrid)
            config: Configuration for the retriever
            index: Index instance (must be FeastIndex)
            format_document: Optional function to format retrieved documents
            id_field: Field to use as document ID
            text_field: Field to use as text field name
            **kwargs: Additional arguments passed to RagRetriever
        """
        if search_type.lower() not in self.VALID_SEARCH_TYPES:
            raise ValueError(
                f"Unsupported search_type {search_type}. "
                f"Must be one of: {self.VALID_SEARCH_TYPES}"
            )

        # move to gpu if available
        torch = get_torch()
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.question_encoder = question_encoder.to(self.device)  # type: ignore
        self.generator_model = generator_model.to(self.device)  # type: ignore

        self.question_encoder_tokenizer = question_encoder_tokenizer
        self.generator_tokenizer = generator_tokenizer

        super().__init__(
            config=config,
            question_encoder_tokenizer=self.question_encoder_tokenizer,
            generator_tokenizer=self.generator_tokenizer,
            index=index,
            init_retrieval=False,
            **kwargs,
        )

        self.feast_repo_path = feast_repo_path
        self.search_type = search_type.lower()
        self.format_document = format_document or self._default_format_document
        self.id_field = id_field
        self.text_field = text_field
        self.feature_view = feature_view
        self.features = features

        # Initialize these lazily for compatability with distributed training/tuning
        self._feature_store = None
        self._vector_store = None

    @property
    def feature_store(self):
        # Initialize FeatureStore lazily
        if self._feature_store is None:
            self._feature_store = FeatureStore(repo_path=self.feast_repo_path)
        return self._feature_store

    @property
    def vector_store(self):
        # Initialize FeastVectorStore lazily
        if self._vector_store is None:
            self._vector_store = FeastVectorStore(
                repo_path=self.feast_repo_path,
                rag_view=self.feature_view,
                features=self.features,
            )
        return self._vector_store

    def retrieve(  # type: ignore [override]
        self,
        question_hidden_states: np.ndarray,
        n_docs: int,
        query: Optional[str] = None,
    ) -> tuple[np.ndarray, np.ndarray, list[dict]]:
        """
        Overrides the base `retrieve` method to fetch documents from Feast.

        This method processes a batch of questions, queries the vector store,
        and formats the results into the 3-part tuple expected by the RAG model:
        (document_embeddings, document_ids, document_dictionaries).

        Args:
            question_hidden_states (np.ndarray):
                Hidden state representation of the question from the encoder.
                Expected shape is (batch_size, seq_len, hidden_dim).
            n_docs (int):
                Number of top documents to retrieve.
            query (Optional[str]):
                Optional raw query string. If not provided and search_type is "text" or "hybrid",
                it will be decoded from question_hidden_states.

        Returns:
            Tuple containing:
                - retrieved_doc_embeds (np.ndarray):
                    Embeddings of the retrieved documents with shape (batch_size, n_docs, embed_dim).
                - doc_ids (np.ndarray):
                    Array of document IDs or passage identifiers with shape (batch_size, n_docs).
                - doc_dicts (list[dict]):
                    List of dictionaries containing document metadata and text.
                    Each dictionary has keys "text", "id", and "title".
        """
        batch_size = question_hidden_states.shape[0]

        # Convert the question hidden states into a list of 1D query vectors.
        pooled_query_vectors = []
        for i in range(batch_size):
            pooled = question_hidden_states[i]
            # Perform normalization to create a unit vector.
            norm = np.linalg.norm(pooled)
            if norm > 0:
                pooled = pooled / norm
            pooled_query_vectors.append(pooled)

        # Determine embedding dimension for padding
        emb_dim = (
            pooled_query_vectors[0].shape[-1]
            if pooled_query_vectors and pooled_query_vectors[0] is not None
            else self.config.retrieval_vector_size
        )

        # Retrieve documents for each query in batch
        batch_embeddings, batch_doc_ids, batch_metadata = [], [], []

        for i in range(batch_size):
            query_vector = pooled_query_vectors[i]
            if isinstance(query, list):
                query_text = query[i] if i < len(query) else None
            else:
                query_text = query

            # Query Feast to get the raw document data
            response = self.vector_store.query(
                query_vector=query_vector if self.search_type != "text" else None,
                query_string=query_text if self.search_type != "vector" else None,
                top_k=n_docs,
            )
            results_dict = response.to_dict()
            # Dynamically get data using the configured feature names
            texts = results_dict.get(self.text_field, [])
            ids = results_dict.get(self.id_field, [])
            num_retrieved = len(texts)

            embeddings = []
            embedding_field_name = None
            for field in self.feature_view.schema:
                if hasattr(field, "vector_index") and field.vector_index:
                    embedding_field_name = field.name
                    break

            if embedding_field_name:
                embeddings = results_dict.get(embedding_field_name, [])
            else:
                raise ValueError(
                    "Warning: No field with 'vector_index=True' found in schema, cannot extract embeddings"
                )

            # Initialize lists for the current query's results
            current_query_embeddings, current_query_ids, current_query_texts = (
                [],
                [],
                [],
            )

            # Process retrieved documents up to n_docs
            emb_array = np.array([])
            if num_retrieved > 0:
                emb_array = np.array(
                    [np.asarray(emb, dtype=np.float32) for emb in embeddings]
                )
                if emb_array.ndim == 1:  # Reshape if it's a flat array
                    emb_array = emb_array.reshape(num_retrieved, -1)

            for k in range(n_docs):
                if k < num_retrieved:
                    # Append actual retrieved data
                    current_query_texts.append(texts[k])
                    current_query_embeddings.append(emb_array[k])
                    try:
                        current_query_ids.append(int(ids[k]))
                    except (ValueError, TypeError):
                        current_query_ids.append(-1)  # Use placeholder for invalid IDs
                else:
                    # Pad with empty/zero values if fewer than n_docs were found
                    current_query_texts.append("")
                    current_query_embeddings.append(np.zeros(emb_dim, dtype=np.float32))
                    current_query_ids.append(-1)

            # Collate results for the current query
            batch_embeddings.append(np.array(current_query_embeddings))
            batch_doc_ids.append(np.array(current_query_ids))
            batch_metadata.append(
                {
                    "text": current_query_texts,
                    "id": current_query_ids,
                    "title": [""]
                    * n_docs,  # RAG model expects a "title" key during Forward pass
                }
            )

        # Return the Collated Batch Results
        # Convert lists of arrays into single batch arrays.
        return (
            np.array(batch_embeddings, dtype=np.float32),
            np.array(batch_doc_ids, dtype=np.int64),
            batch_metadata,
        )

    def generate_answer(
        self, query: str, top_k: int = 5, max_new_tokens: int = 100
    ) -> str:
        """A helper method to generate an answer for a single query string.

        Args:
            query: The query to answer
            top_k: Number of documents to retrieve
            max_new_tokens: Maximum number of tokens to generate

        Returns:
            Generated answer string
        """
        if not self.question_encoder or not self.generator_model:
            raise ValueError(
                "`question_encoder` and `generator_model` must be provided to use `generate_answer`."
            )
        torch = get_torch()
        inputs = self.question_encoder_tokenizer(query, return_tensors="pt").to(
            self.question_encoder.device
        )
        question_embeddings = self.question_encoder(**inputs).pooler_output
        question_embeddings = (
            question_embeddings.detach().cpu().to(torch.float32).numpy()
        )
        _, _, doc_batch = self.retrieve(question_embeddings, n_docs=top_k, query=query)

        contexts = doc_batch[0]["text"] if doc_batch else []
        context_str = "\n\n".join(filter(None, contexts))

        prompt = f"Context: {context_str}\n\nQuestion: {query}\n\nAnswer:"

        generator_inputs = self.generator_tokenizer(prompt, return_tensors="pt").to(
            self.generator_model.device
        )
        output_ids = self.generator_model.generate(
            **generator_inputs, max_new_tokens=max_new_tokens
        )

        return self.generator_tokenizer.decode(output_ids[0], skip_special_tokens=True)

    def _default_format_document(self, doc: dict[str, Any]) -> str:
        """Default document formatting function.

        Args:
            doc: Document dictionary to format

        Returns:
            Formatted document string
        """
        lines = []
        for key, value in doc.items():
            # Skip vectors by checking for long float lists
            if (
                isinstance(value, list)
                and len(value) > 10
                and all(isinstance(x, (float, int)) for x in value)
            ):
                continue
            lines.append(f"{key.replace('_', ' ').capitalize()}: {value}")
        return "\n".join(lines)



from typing import Literal, Optional

from pydantic import StrictBool, StrictStr

from feast.infra.feature_servers.base_config import BaseFeatureServerConfig


class McpFeatureServerConfig(BaseFeatureServerConfig):
    """MCP (Model Context Protocol) Feature Server configuration."""

    # Feature server type selector
    type: Literal["mcp"] = "mcp"

    # Enable MCP server support - defaults to False as requested
    mcp_enabled: StrictBool = False

    # MCP server name for identification
    mcp_server_name: StrictStr = "feast-mcp-server"

    # MCP server version
    mcp_server_version: StrictStr = "1.0.0"

    # Optional MCP transport configuration
    mcp_transport: Optional[StrictStr] = None

    # The endpoint definition for transformation_service (inherited from base)
    transformation_service_endpoint: StrictStr = "localhost:6566"



import logging
import threading
from concurrent import futures
from typing import Optional, Union

import grpc
import pandas as pd
from grpc_health.v1 import health, health_pb2_grpc

from feast.data_source import PushMode
from feast.errors import FeatureServiceNotFoundException, PushSourceNotFoundException
from feast.feature_service import FeatureService
from feast.feature_store import FeatureStore
from feast.protos.feast.serving.GrpcServer_pb2 import (
    PushResponse,
    WriteToOnlineStoreResponse,
)
from feast.protos.feast.serving.GrpcServer_pb2_grpc import (
    GrpcFeatureServerServicer,
    add_GrpcFeatureServerServicer_to_server,
)
from feast.protos.feast.serving.ServingService_pb2 import (
    GetOnlineFeaturesRequest,
    GetOnlineFeaturesResponse,
)

logger = logging.getLogger(__name__)


def parse(features):
    df = {}
    for i in features.keys():
        df[i] = [features.get(i)]
    return pd.DataFrame.from_dict(df)


class GrpcFeatureServer(GrpcFeatureServerServicer):
    fs: FeatureStore

    _shuting_down: bool = False
    _active_timer: Optional[threading.Timer] = None

    def __init__(self, fs: FeatureStore, registry_ttl_sec: int = 5):
        self.fs = fs
        self.registry_ttl_sec = registry_ttl_sec
        super().__init__()

        self._async_refresh()

    def Push(self, request, context):
        try:
            df = parse(request.features)
            if request.to == "offline":
                to = PushMode.OFFLINE
            elif request.to == "online":
                to = PushMode.ONLINE
            elif request.to == "online_and_offline":
                to = PushMode.ONLINE_AND_OFFLINE
            else:
                raise ValueError(
                    f"{request.to} is not a supported push format. Please specify one of these ['online', 'offline', "
                    f"'online_and_offline']."
                )
            self.fs.push(
                push_source_name=request.push_source_name,
                df=df,
                allow_registry_cache=request.allow_registry_cache,
                to=to,
            )
        except PushSourceNotFoundException as e:
            logger.exception(str(e))
            context.set_code(grpc.StatusCode.INVALID_ARGUMENT)
            context.set_details(str(e))
            return PushResponse(status=False)
        except Exception as e:
            logger.exception(str(e))
            context.set_code(grpc.StatusCode.INTERNAL)
            context.set_details(str(e))
            return PushResponse(status=False)
        return PushResponse(status=True)

    def WriteToOnlineStore(self, request, context):
        logger.warning(
            "write_to_online_store is deprecated. Please consider using Push instead"
        )
        try:
            df = parse(request.features)
            self.fs.write_to_online_store(
                feature_view_name=request.feature_view_name,
                df=df,
                allow_registry_cache=request.allow_registry_cache,
            )
        except Exception as e:
            logger.exception(str(e))
            context.set_code(grpc.StatusCode.INTERNAL)
            context.set_details(str(e))
            return PushResponse(status=False)
        return WriteToOnlineStoreResponse(status=True)

    def GetOnlineFeatures(self, request: GetOnlineFeaturesRequest, context):
        if request.HasField("feature_service"):
            logger.info(f"Requesting feature service: {request.feature_service}")
            try:
                features: Union[list[str], FeatureService] = (
                    self.fs.get_feature_service(
                        request.feature_service, allow_cache=True
                    )
                )
            except FeatureServiceNotFoundException as e:
                logger.error(f"Feature service {request.feature_service} not found")
                context.set_code(grpc.StatusCode.INTERNAL)
                context.set_details(str(e))
                return GetOnlineFeaturesResponse()
        else:
            features = list(request.features.val)

        result = self.fs.get_online_features(
            features,
            request.entities,
            request.full_feature_names,
        ).proto

        return result

    def _async_refresh(self):
        self.fs.refresh_registry()
        if self._shuting_down:
            return
        self._active_timer = threading.Timer(self.registry_ttl_sec, self._async_refresh)
        self._active_timer.start()


def get_grpc_server(
    address: str,
    fs: FeatureStore,
    max_workers: int,
    registry_ttl_sec: int,
):
    logger.info(f"Initializing gRPC server on {address}")
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
    add_GrpcFeatureServerServicer_to_server(
        GrpcFeatureServer(fs, registry_ttl_sec=registry_ttl_sec),
        server,
    )
    health_servicer = health.HealthServicer(
        experimental_non_blocking=True,
        experimental_thread_pool=futures.ThreadPoolExecutor(max_workers=max_workers),
    )
    health_pb2_grpc.add_HealthServicer_to_server(health_servicer, server)
    server.add_insecure_port(address)
    return server



from feast.permissions.auth_model import (
    AuthConfig,
)
from feast.permissions.client.auth_client_manager import (
    AuthenticationClientManagerFactory,
)


def get_auth_token(auth_config: AuthConfig) -> str:
    return (
        AuthenticationClientManagerFactory(auth_config)
        .get_auth_client_manager()
        .get_token()
    )



import json
from typing import Any, Callable, Iterable, Optional, Tuple, Type

from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.datatypes.container import Array
from clickhouse_connect.datatypes.numeric import (
    Boolean,
    Decimal,
    Float32,
    Float64,
    Int32,
    Int64,
)
from clickhouse_connect.datatypes.registry import get_from_name
from clickhouse_connect.datatypes.string import String
from clickhouse_connect.datatypes.temporal import DateTime, DateTime64

from feast import RepoConfig, ValueType
from feast.data_source import DataSource
from feast.errors import DataSourceNoNameException
from feast.infra.utils.clickhouse.connection_utils import get_client
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.protos.feast.core.SavedDataset_pb2 import (
    SavedDatasetStorage as SavedDatasetStorageProto,
)
from feast.saved_dataset import SavedDatasetStorage


class ClickhouseOptions:
    def __init__(
        self,
        name: Optional[str],
        query: Optional[str],
        table: Optional[str],
    ):
        self._name = name or ""
        self._query = query or ""
        self._table = table or ""

    @classmethod
    def from_proto(cls, clickhouse_options_proto: DataSourceProto.CustomSourceOptions):
        config = json.loads(clickhouse_options_proto.configuration.decode("utf8"))
        postgres_options = cls(
            name=config["name"], query=config["query"], table=config["table"]
        )

        return postgres_options

    def to_proto(self) -> DataSourceProto.CustomSourceOptions:
        clickhouse_options_proto = DataSourceProto.CustomSourceOptions(
            configuration=json.dumps(
                {"name": self._name, "query": self._query, "table": self._table}
            ).encode()
        )
        return clickhouse_options_proto


class ClickhouseSource(DataSource):
    def source_type(self) -> DataSourceProto.SourceType.ValueType:
        # TODO: Add ClickhouseSourceType to DataSourceProto
        return DataSourceProto.CUSTOM_SOURCE

    def __init__(
        self,
        name: Optional[str] = None,
        query: Optional[str] = None,
        table: Optional[str] = None,
        timestamp_field: Optional[str] = "",
        created_timestamp_column: Optional[str] = "",
        field_mapping: Optional[dict[str, str]] = None,
        description: Optional[str] = "",
        tags: Optional[dict[str, str]] = None,
        owner: Optional[str] = "",
    ):
        self._clickhouse_options = ClickhouseOptions(
            name=name, query=query, table=table
        )

        if name is None and table is None:
            raise DataSourceNoNameException()
        name = name or table
        assert name

        super().__init__(
            name=name,
            timestamp_field=timestamp_field,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            description=description,
            tags=tags,
            owner=owner,
        )

    @staticmethod
    def from_proto(data_source: DataSourceProto) -> Any:
        assert data_source.HasField("custom_options")

        postgres_options = json.loads(data_source.custom_options.configuration)

        return ClickhouseSource(
            name=postgres_options["name"],
            query=postgres_options["query"],
            table=postgres_options["table"],
            field_mapping=dict(data_source.field_mapping),
            timestamp_field=data_source.timestamp_field,
            created_timestamp_column=data_source.created_timestamp_column,
            description=data_source.description,
            tags=dict(data_source.tags),
            owner=data_source.owner,
        )

    def _to_proto_impl(self) -> DataSourceProto:
        data_source_proto = DataSourceProto(
            name=self.name,
            type=DataSourceProto.CUSTOM_SOURCE,
            data_source_class_type="feast.infra.offline_stores.contrib.clickhouse_offline_store.clickhouse_source.ClickhouseSource",
            field_mapping=self.field_mapping,
            custom_options=self._clickhouse_options.to_proto(),
            description=self.description,
            tags=self.tags,
            owner=self.owner,
        )

        data_source_proto.timestamp_field = self.timestamp_field
        data_source_proto.created_timestamp_column = self.created_timestamp_column

        return data_source_proto

    def validate(self, config: RepoConfig):
        pass

    @staticmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        return ch_type_to_feast_value_type

    def get_table_column_names_and_types(
        self, config: RepoConfig
    ) -> Iterable[Tuple[str, str]]:
        with get_client(config.offline_store) as client:
            result = client.query(
                f"SELECT * FROM {self.get_table_query_string()} AS sub LIMIT 0"
            )
            column_types = list(zip(result.column_names, result.column_types))
            return [
                (name, _ch_type_to_ch_type_str(type_)) for name, type_ in column_types
            ]

    def get_table_query_string(self) -> str:
        if self._clickhouse_options._table:
            return f"{self._clickhouse_options._table}"
        else:
            return f"({self._clickhouse_options._query})"


class SavedDatasetClickhouseStorage(SavedDatasetStorage):
    _proto_attr_name = "custom_storage"

    clickhouse_options: ClickhouseOptions

    def __init__(self, table_ref: str):
        self.clickhouse_options = ClickhouseOptions(
            table=table_ref, name=None, query=None
        )

    @staticmethod
    def from_proto(storage_proto: SavedDatasetStorageProto) -> SavedDatasetStorage:
        return SavedDatasetClickhouseStorage(
            table_ref=ClickhouseOptions.from_proto(storage_proto.custom_storage)._table
        )

    def to_proto(self) -> SavedDatasetStorageProto:
        return SavedDatasetStorageProto(
            custom_storage=self.clickhouse_options.to_proto()
        )

    def to_data_source(self) -> DataSource:
        return ClickhouseSource(table=self.clickhouse_options._table)


def ch_type_to_feast_value_type(type_str: str) -> ValueType:
    type_obj = get_from_name(type_str)
    type_cls = type(type_obj)
    container_type = None
    if isinstance(type_obj, Array):
        container_type = Array
    type_map: dict[
        tuple[Optional[Type[ClickHouseType]], Type[ClickHouseType]], ValueType
    ] = {
        (None, Boolean): ValueType.BOOL,
        (None, String): ValueType.STRING,
        (None, Float32): ValueType.FLOAT,
        (None, Float64): ValueType.DOUBLE,
        (None, Decimal): ValueType.DOUBLE,
        (None, Int32): ValueType.INT32,
        (None, Int64): ValueType.INT64,
        (None, DateTime): ValueType.UNIX_TIMESTAMP,
        (None, DateTime64): ValueType.UNIX_TIMESTAMP,
        (Array, Boolean): ValueType.BOOL_LIST,
        (Array, String): ValueType.STRING_LIST,
        (Array, Float32): ValueType.FLOAT_LIST,
        (Array, Float64): ValueType.DOUBLE_LIST,
        (Array, Decimal): ValueType.DOUBLE_LIST,
        (Array, Int32): ValueType.INT32_LIST,
        (Array, Int64): ValueType.INT64_LIST,
        (Array, DateTime): ValueType.UNIX_TIMESTAMP_LIST,
        (Array, DateTime64): ValueType.UNIX_TIMESTAMP_LIST,
    }
    value = type_map.get((container_type, type_cls), ValueType.UNKNOWN)
    if value == ValueType.UNKNOWN:
        print("unknown type:", type_str)
    return value


def _ch_type_to_ch_type_str(type_: ClickHouseType) -> str:
    return type_.name



#!/usr/bin/env python3

"""
Test workflow for Ray offline store and compute engine template.

This script demonstrates:
1. Ray offline store for efficient data I/O
2. Ray compute engine for distributed feature processing
3. Historical feature retrieval with point-in-time joins
4. Feature materialization to online store
5. Online feature serving

Run this after: feast apply
"""

import sys
from datetime import datetime, timedelta
from pathlib import Path

import pandas as pd

# Add the feature repo to the path
repo_path = Path(__file__).parent
sys.path.append(str(repo_path))

try:
    from feast import FeatureStore
except ImportError:
    print("Please install feast: pip install feast[ray]")
    sys.exit(1)


def run_demo():
    print("=" * 60)
    print("🚀 Ray Offline Store & Compute Engine Demo")
    print("=" * 60)

    # Initialize the feature store
    print("\n1. Initializing Feast with Ray configuration...")
    store = FeatureStore(repo_path=".")

    print(f"   ✓ Offline store: {store.config.offline_store.type}")
    if hasattr(store.config, "batch_engine") and store.config.batch_engine:
        print(f"   ✓ Compute engine: {store.config.batch_engine.type}")
    else:
        print("   ⚠ No compute engine configured")

    # Create entity DataFrame for historical features
    print("\n2. Creating entity DataFrame for historical feature retrieval...")
    end_date = datetime.now().replace(microsecond=0, second=0, minute=0)
    start_date = end_date - timedelta(days=2)

    entity_df = pd.DataFrame(
        {
            "driver_id": [1001, 1002, 1003],
            "customer_id": [2001, 2002, 2003],
            "event_timestamp": [
                pd.Timestamp(end_date - timedelta(hours=24), tz="UTC"),
                pd.Timestamp(end_date - timedelta(hours=12), tz="UTC"),
                pd.Timestamp(end_date - timedelta(hours=6), tz="UTC"),
            ],
        }
    )

    print(f"   ✓ Created entity DataFrame with {len(entity_df)} rows")
    print(f"   ✓ Time range: {start_date} to {end_date}")

    # Retrieve historical features using Ray compute engine
    print("\n3. Retrieving historical features with Ray compute engine...")
    print("   (This demonstrates distributed point-in-time joins)")

    try:
        # Get historical features - this uses Ray compute engine for distributed processing
        historical_features = store.get_historical_features(
            entity_df=entity_df,
            features=[
                "driver_hourly_stats:conv_rate",
                "driver_hourly_stats:acc_rate",
                "driver_hourly_stats:avg_daily_trips",
                "customer_daily_profile:current_balance",
                "customer_daily_profile:avg_passenger_count",
                "customer_daily_profile:lifetime_trip_count",
            ],
        )

        # Convert to DataFrame - Ray processes this efficiently
        historical_df = historical_features.to_df()
        print(f"   ✓ Retrieved {len(historical_df)} historical feature rows")
        print(f"   ✓ Features: {list(historical_df.columns)}")

        # Show sample of the data
        print("\n   Sample historical features:")
        print(historical_df.head(3).to_string(index=False))

    except Exception as e:
        print(f"   ⚠ Historical features retrieval failed: {e}")
        print("   This might be due to missing Ray dependencies or data")

    # Demonstrate on-demand feature transformations
    print("\n4. Testing on-demand feature transformations...")
    try:
        # Get features including on-demand transformations
        features_with_odfv = store.get_historical_features(
            entity_df=entity_df.head(1),
            features=[
                "driver_hourly_stats:conv_rate",
                "driver_hourly_stats:acc_rate",
                "driver_hourly_stats:avg_daily_trips",
                "driver_activity_v2:conv_rate_plus_acc_rate",
                "driver_activity_v2:trips_per_day_normalized",
            ],
        )

        odfv_df = features_with_odfv.to_df()
        print(f"   ✓ Retrieved {len(odfv_df)} rows with on-demand transformations")

        # Show sample with transformations
        print("\n   Sample with on-demand features:")
        print(
            odfv_df[["driver_id", "conv_rate", "acc_rate", "conv_rate_plus_acc_rate"]]
            .head(3)
            .to_string(index=False)
        )

    except Exception as e:
        print(f"   ⚠ On-demand features failed: {e}")

    # Materialize features to online store
    print("\n5. Materializing features to online store...")
    try:
        materialize_end = end_date

        print(f"   Attempting materialization up to {materialize_end}")

        # Try materialization with Ray compute engine
        store.materialize_incremental(end_date=materialize_end)
        print("   ✓ Ray compute engine materialization successful!")

        # Test online feature retrieval
        print("\n6. Testing online feature serving...")
        online_features = store.get_online_features(
            features=[
                "driver_hourly_stats:conv_rate",
                "driver_hourly_stats:acc_rate",
                "customer_daily_profile:current_balance",
            ],
            entity_rows=[
                {"driver_id": 1001, "customer_id": 2001},
                {"driver_id": 1002, "customer_id": 2002},
            ],
        )

        online_df = online_features.to_df()
        print(f"   ✓ Retrieved {len(online_df)} online feature rows")
        print("\n   Sample online features:")
        print(online_df.to_string(index=False))

    except Exception as e:
        print(f"   ⚠ Materialization/online serving failed: {e}")

    print("\n" + "=" * 60)
    print("🎉 Ray Demo Complete!")
    print("=" * 60)

    print(
        "\nIf you want to explore Feast with your existing ray cluster, you can configure ray_address to feature_store.yaml: \n"
    )
    print("""
    offline_store:
      ray_address: "127.0.0.1:10001"
    batch_engine:
      ray_address: "127.0.0.1:10001"
    """)


if __name__ == "__main__":
    run_demo()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/Registry.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from feast.protos.feast.core import Entity_pb2 as feast_dot_core_dot_Entity__pb2
from feast.protos.feast.core import FeatureService_pb2 as feast_dot_core_dot_FeatureService__pb2
from feast.protos.feast.core import FeatureTable_pb2 as feast_dot_core_dot_FeatureTable__pb2
from feast.protos.feast.core import FeatureView_pb2 as feast_dot_core_dot_FeatureView__pb2
from feast.protos.feast.core import InfraObject_pb2 as feast_dot_core_dot_InfraObject__pb2
from feast.protos.feast.core import OnDemandFeatureView_pb2 as feast_dot_core_dot_OnDemandFeatureView__pb2
from feast.protos.feast.core import StreamFeatureView_pb2 as feast_dot_core_dot_StreamFeatureView__pb2
from feast.protos.feast.core import DataSource_pb2 as feast_dot_core_dot_DataSource__pb2
from feast.protos.feast.core import SavedDataset_pb2 as feast_dot_core_dot_SavedDataset__pb2
from feast.protos.feast.core import ValidationProfile_pb2 as feast_dot_core_dot_ValidationProfile__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from feast.protos.feast.core import Permission_pb2 as feast_dot_core_dot_Permission__pb2
from feast.protos.feast.core import Project_pb2 as feast_dot_core_dot_Project__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x66\x65\x61st/core/Registry.proto\x12\nfeast.core\x1a\x17\x66\x65\x61st/core/Entity.proto\x1a\x1f\x66\x65\x61st/core/FeatureService.proto\x1a\x1d\x66\x65\x61st/core/FeatureTable.proto\x1a\x1c\x66\x65\x61st/core/FeatureView.proto\x1a\x1c\x66\x65\x61st/core/InfraObject.proto\x1a$feast/core/OnDemandFeatureView.proto\x1a\"feast/core/StreamFeatureView.proto\x1a\x1b\x66\x65\x61st/core/DataSource.proto\x1a\x1d\x66\x65\x61st/core/SavedDataset.proto\x1a\"feast/core/ValidationProfile.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1b\x66\x65\x61st/core/Permission.proto\x1a\x18\x66\x65\x61st/core/Project.proto\"\xff\x05\n\x08Registry\x12$\n\x08\x65ntities\x18\x01 \x03(\x0b\x32\x12.feast.core.Entity\x12\x30\n\x0e\x66\x65\x61ture_tables\x18\x02 \x03(\x0b\x32\x18.feast.core.FeatureTable\x12.\n\rfeature_views\x18\x06 \x03(\x0b\x32\x17.feast.core.FeatureView\x12,\n\x0c\x64\x61ta_sources\x18\x0c \x03(\x0b\x32\x16.feast.core.DataSource\x12@\n\x17on_demand_feature_views\x18\x08 \x03(\x0b\x32\x1f.feast.core.OnDemandFeatureView\x12;\n\x14stream_feature_views\x18\x0e \x03(\x0b\x32\x1d.feast.core.StreamFeatureView\x12\x34\n\x10\x66\x65\x61ture_services\x18\x07 \x03(\x0b\x32\x1a.feast.core.FeatureService\x12\x30\n\x0esaved_datasets\x18\x0b \x03(\x0b\x32\x18.feast.core.SavedDataset\x12>\n\x15validation_references\x18\r \x03(\x0b\x32\x1f.feast.core.ValidationReference\x12 \n\x05infra\x18\n \x01(\x0b\x32\x11.feast.core.Infra\x12\x39\n\x10project_metadata\x18\x0f \x03(\x0b\x32\x1b.feast.core.ProjectMetadataB\x02\x18\x01\x12\x1f\n\x17registry_schema_version\x18\x03 \x01(\t\x12\x12\n\nversion_id\x18\x04 \x01(\t\x12\x30\n\x0clast_updated\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12+\n\x0bpermissions\x18\x10 \x03(\x0b\x32\x16.feast.core.Permission\x12%\n\x08projects\x18\x11 \x03(\x0b\x32\x13.feast.core.Project\"8\n\x0fProjectMetadata\x12\x0f\n\x07project\x18\x01 \x01(\t\x12\x14\n\x0cproject_uuid\x18\x02 \x01(\tBR\n\x10\x66\x65\x61st.proto.coreB\rRegistryProtoZ/github.com/feast-dev/feast/go/protos/feast/coreb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.core.Registry_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020feast.proto.coreB\rRegistryProtoZ/github.com/feast-dev/feast/go/protos/feast/core'
  _globals['_REGISTRY'].fields_by_name['project_metadata']._options = None
  _globals['_REGISTRY'].fields_by_name['project_metadata']._serialized_options = b'\030\001'
  _globals['_REGISTRY']._serialized_start=449
  _globals['_REGISTRY']._serialized_end=1216
  _globals['_PROJECTMETADATA']._serialized_start=1218
  _globals['_PROJECTMETADATA']._serialized_end=1274
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

* Copyright 2020 The Feast Authors
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     https://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Project(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SPEC_FIELD_NUMBER: builtins.int
    META_FIELD_NUMBER: builtins.int
    @property
    def spec(self) -> global___ProjectSpec:
        """User-specified specifications of this entity."""
    @property
    def meta(self) -> global___ProjectMeta:
        """System-populated metadata for this entity."""
    def __init__(
        self,
        *,
        spec: global___ProjectSpec | None = ...,
        meta: global___ProjectMeta | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["meta", b"meta", "spec", b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["meta", b"meta", "spec", b"spec"]) -> None: ...

global___Project = Project

class ProjectSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class TagsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    TAGS_FIELD_NUMBER: builtins.int
    OWNER_FIELD_NUMBER: builtins.int
    name: builtins.str
    """Name of the Project"""
    description: builtins.str
    """Description of the Project"""
    @property
    def tags(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """User defined metadata"""
    owner: builtins.str
    """Owner of the Project"""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        description: builtins.str = ...,
        tags: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        owner: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["description", b"description", "name", b"name", "owner", b"owner", "tags", b"tags"]) -> None: ...

global___ProjectSpec = ProjectSpec

class ProjectMeta(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREATED_TIMESTAMP_FIELD_NUMBER: builtins.int
    LAST_UPDATED_TIMESTAMP_FIELD_NUMBER: builtins.int
    @property
    def created_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Time when the Project is created"""
    @property
    def last_updated_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Time when the Project is last updated with registry changes (Apply stage)"""
    def __init__(
        self,
        *,
        created_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        last_updated_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["created_timestamp", b"created_timestamp", "last_updated_timestamp", b"last_updated_timestamp"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["created_timestamp", b"created_timestamp", "last_updated_timestamp", b"last_updated_timestamp"]) -> None: ...

global___ProjectMeta = ProjectMeta
